
//Project:      mosaic
//Version:      0.3
//Copyright:    Copyright (c) 1998
//Author:       Herbert Praehofer, Johannes Sametinger, Alois Stritzinger
//Affiliation:  Johannes Kepler University, Linz, Austria

// -----------------------------------------------------------------------------
// Class BasicActiveModel      -------------------------------------------------
// -----------------------------------------------------------------------------
// programmed by: hp
// last modified by: hp, 14.12.98
// documented by: hp, 7.4.99

package mosaic.sim.kernel;

import mosaic.sim.kernel.control.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;
import mosaic.sim.kernel.variable.*;
import mosaic.sim.kernel.simevent.*;
import mosaic.sim.util.artificialidentifier.ArtificialIDGiver;

/**
* Class implementing a root class for active model implementation.
* Extends BasicModel and additionally implements TimeEventComponent, that means,
* BasicActiveModel can be scheduled for time events and can process time events.
* BasicActiveModel should therefore seen as combination of a BasicModel with one Timer.
* <p>
* To extend this class and eventually implement your active models based on it
* specialize the following methods as follows:
* <ul>
* <li> <code>processSimEvent</code> to react to time events.
* </ul>
* <p>
* Use the following methods to schedule the active model for time events :
* <ul>
* <li> <code>activateIn (deltaT) </code> for a time event in deltaT time units.
* <li> <code>activateAt (t) </code> for a time event at t (be careful that t is not in the past!!).
* <li> <code>activate () </code> for a time event to occur immediately.
* <li> <code>passivate () </code> to cancel a scheduled time event.
* <li> <code>setSelectOrder </code> to define the priority of this model in case of simultaneous events.
* </ul>
* <p>
* @see sim.kernel.ModelComponent
* @see sim.kernel.Timer
* @see sim.kernel.BasicModel
*/
public class BasicActiveModel extends BasicModel implements TimeEventComponent, TimeEventListener  {


// Constructors ----------------------------------------------------------------

  /** Default constructor */
  public BasicActiveModel() {
	this ("active model");
  }

  /**  Constructor with parent model and label. */
  public BasicActiveModel(String label) {
	super(label);
	try  {
	  jbInit();
	}
	catch (Exception e) {
	  e.printStackTrace(getSimApplication().getLogPrinter().getLogStream("error"));
	}
  }

// initialize ------------------------------------------------------------------

  public void initialize() {
	super.initialize();
	activateAt(getInitTime());
  }


// -----------------------------------------------------------------------------------------------------------
// interface ArtificalIDOwner --------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------
  private final int artificialID=ArtificialIDGiver.getID();
  public final int getArtificialID() {
	return artificialID;
  }

// -----------------------------------------------------------------------------
// interface Comparable --------------------------------------------------------
// -----------------------------------------------------------------------------

  /** compares this timer with another timer c based on tn and selectOrder */
  public boolean before(Comparable c) {
	if (c instanceof TimeEventComponent) {
	  TimeEventComponent s= (TimeEventComponent)c;
	  if (getTN()==s.getTN())
		return getSelectOrder()>s.getSelectOrder();
	  else
		return getTN()<=s.getTN();
	}
	return false;
  }

// -----------------------------------------------------------------------------
// interface TimeEventComponent ---------------------------------------------------
// -----------------------------------------------------------------------------

// next event time -------------------------------------------------------------

  /** variable to hold the time tn of the next event for the timer */
  protected double tn= 0;

  /** Gets the time tn of the next event for the timer */
  public double getTN() {
	return tn;
  }

  /** Sets the time tn of the next event for the timer */
  public void setTN(double newTime) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	double oldTN = tn;
	tn = newTime;
  }


// last event time -------------------------------------------------------------

  /** variable to hold the time tl of the last event for the timer */
  protected double tl= 0;

  /** Gets the time tl of the last event for the timer */
  public double getTL() { return tl; }

  /** Sets the time tl of the last event for the timer */
  public void setTL(double newTime) {
	double oldTL = tn;
	tl = newTime;
  }


// select order  ---------------------------------------------------------------

  /** variable to hold the selectOrder for the timer which is used as a priority
  * for tie breaking between several timer with equal next event time.
  */
  protected int selectOrder = 0;

  /** Gets the select order of the timer */
  public int getSelectOrder() {
	  return selectOrder;
  }

  /** Sets the select order of the timer */
  public void setSelectOrder(int newOrder) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	int oldSelectOrder = selectOrder;
	selectOrder= newOrder;
  }

// triggerEvent  ---------------------------------------------------------------

  /** Triggers an event, that is, it executes the internal transition function */
	public void triggerEvent() {
	  setTL(getTime());
	  unregisterEvent();
	  timeEvent.setTime(getTime());
	  Vector timeEventListenersCopy = (Vector)(timeEventListeners.clone());
	  for (int idx = 0; idx < timeEventListenersCopy.size(); idx++) {
		  ((SimEventListener)timeEventListenersCopy.get(idx)).processSimEvent (timeEvent);
	  }
	  processSimEvent(timeEvent);
  }


// timeEvent -------------------------------------------------------------------

  /** Vector of TimeEventListeners */
  Vector timeEventListeners = new Vector();

  /** single TimeEvent in Timer */
  TimeEvent timeEvent = new TimeEvent(this);

  /** Add SimEventListener */
  public void addSimEventListener (SimEventListener l) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	  timeEventListeners.add(l);
	  l.addSimEventComponent(this);
  }

  /** Remove SimEventListener */
  public void removeSimEventListener (SimEventListener l){
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	  timeEventListeners.remove(l);
	  l.removeSimEventComponent(this);
  }

  /** Add TimeEventListener */
	public void addTimeEventListener (TimeEventListener l) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	  addSimEventListener(l);
  }

  /** Remove TimeEventListener */
  public void removeTimeEventListener (TimeEventListener l){
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	  removeSimEventListener(l);
  }

// -----------------------------------------------------------------------------
// TimeEventListener, SimEventListener  ----------------------------------------
// -----------------------------------------------------------------------------

  /** Adds a TimeEventComponent. Is called from addTimeEventListener of TimeEventComponent. */
  public void addTimeEventComponent (SimEventComponent c){
  }

  /** Removes a TimeEventComponent. Is called from removeTimeEventListener of TimeEventComponent.  */
  public void removeTimeEventComponent (SimEventComponent c){
  }

  /** Reacts to SimEvents from SimEventComponents
  * This is the event method for reacting to SimEvents
  */
  public void processSimEvent (SimEvent evt) {
   // overwrite this method
  }

  /** Adds a SimEventComponent this listener listens to. Is called from addSimEventListener of SimEventComponent. */
  public void addSimEventComponent (SimEventComponent c){
  }

  /** Removes a SimEventComponent this listener listens to. Is called from removeSimEventListener of SimEventComponent.  */
  public void removeSimEventComponent (SimEventComponent c){
  }

// -----------------------------------------------------------------------------
// scheduling functions used in the model specifications -----------------------
// -----------------------------------------------------------------------------

  //** Activates the Timer immediately */
  public void activate (){
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	registerEvent(getTime());
  }

  //** Activates the Timer at time <code>time</code> */
  public void activateAt (double time){
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	registerEvent(time);
  }

  //** Activates the Timer in  <code>deltaT</code> time units */
  public void activateIn (double deltaT){
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	registerEvent(getTime()+deltaT);
  }

  //** Passivates the Timer by removing it from the event list */
  public void passivate(){
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	unregisterEvent ();
  }


// stop and restart  -----------------------------------------------------------
// (a stopped Timer can be restarted again with the remaining time)

  /** variable to hold the remaining time when stopped */
  protected double deltaT = 0;

  /** boolean to signify if the timer has been stopped */
  protected boolean stopped = false;

  /** Stops the timer and saves the remaining time */
  public void stop () {
	  deltaT = getTN() - getTime();
	  setStopped (true);
	  passivate ();
  }

  /** Restarts the timer with remaining time */
  public void restart () {
	if (isStopped()) {
	  setStopped (false);
	  activateIn (deltaT);
	}
  }

  /** Tests if the timer has been stopped */
  public boolean isStopped () {
	return stopped;
  }

  /** Sets the stopped flag of */
  public void setStopped (boolean newStopped) {
	stopped= newStopped;
  }


//property active and passive --------------------------------------------------

  /** variable to hold the property active */
  protected BooleanVariable activeVar = new BooleanVariable("active", false);

  /** Tests if the timer is active */
  public boolean isActive() {
	return activeVar.getValue();
  }

  /** Sets the active property of the timer */
  protected void setActive(boolean a) {
	activeVar.setValue(a);
  }

  /** Tests if the timer is passive (! active) */
  public boolean isPassive() {
	return !activeVar.getValue();
  }

  /** Add a VariableChangeListener for variable active */
  public void addActiveListener (VariableChangeListener l) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	activeVar.addVariableChangeListener(l);
  }

  /** Remove a VariableChangeListener for variable active */
  public void removeActiveListener (VariableChangeListener l) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	activeVar.removeVariableChangeListener(l);
  }



/// property initTime for first simulation step  -------------------------------

  /** variable to hold the initial event time for the timer. Is INF by default.  */
  protected double initTime = INF;

  /** Gets the initial event time for the timer. */
  public double getInitTime() {
	return initTime;
  }

  /** Sets the initial event time for the timer. */
  public void setInitTime(double newInitTime) {
	initTime= newInitTime;
  }


// event registration ----------------------------------------------------------

  /** Registers an event for time <code>time</time>, that is, it is inserted into the event list. */
  protected void registerEvent(double time) {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	if (isStopped()) setStopped (false);
	//setTN(time);
	if (time >= getTime()) {
		setActive (true);
		getSimApplication().getSimulator().scheduleEvent(this, time);
	}
	else {
		printErrorOutput("tried to schedule past event:" + time + "<"
			+ getTime() + " in timer " + toString() + "!");
		unregisterEvent();
	}
  }

  /** Unregisters a Timer, that is, it is removed from the event list. */
  protected void unregisterEvent() {
	if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
	  getSimApplication().getIntegrator().stateEventOccurred = true;
	  return;
	}
	//setTN(INF);
	getSimApplication().getSimulator().unscheduleEvent(this);
	setActive(false);
  }

	/** Tests if the timer is the next in the event list. */
  public boolean isNextToFire() {
	return getSimApplication().getSimulator().isNextToFire(this);
  }


  private void jbInit() throws Exception {
	this.setLayout(null);
  }


} // end BasicActiveModel  -----------------------------------------------------


