/*
 * BasicEventSupport
 *
 * @date 19.4.2000
 *
 * copyright: JKU Linz
 */

package mosaic.sim.kernel;

/**
 * This class implements a basic event support, but it shound not be used directly.
 * Instead this class should used to create specific EventSupport classes using copy
 * and paste.
 * To adjust it: <ul>
 * <li> replace the EventListener class with the specific EventListener class </li>
 * <li> implement event propagation methods, e.g. fireChangeEvent. Base them on the
 *      inform listeners method. </li>
 * <li> specify the correct information method at the EventListener! </li>
 * <li> This version uses a read/write locked HashMap. If the specific EventSupport doesn't
 *      need thread safty, then the wrapper can be removed. </li>
 * </ul>
 * To preserve consistency, any major changes should be made in this file and then
 * be updated at the other EventSupport classes, that are based on this one.
 */

import java.util.*;
import concurrent.CopyOnWriteArrayList;
import concurrent.SyncMap;
import concurrent.ReentrantWriterPreferenceReadWriteLock;

public class BasicEventSupport{

    /** Maps the eventName to the list of listeners. */
    protected Map lists=new SyncMap(new HashMap(15),new ReentrantWriterPreferenceReadWriteLock());

    /** Contains the listeners that should get all events. */
    protected List allEvents=new CopyOnWriteArrayList();

    /** Adds the listener to the list for receiving all events. */
    public void addListenerAllEvents(EventListener listener){
        allEvents.add(listener);
    }

    /** Registers a listener for special events. */
    public void addListener(String eventName, EventListener listener){
        List list=(List) lists.get(eventName);
        if (list==null){ // create new list
            list=new CopyOnWriteArrayList();
            list.add(listener);
            lists.put(eventName,list);
        }else{
            list.add(listener);
        }
    }

    /**
     * This method removes the listener from the list of listeners that receive all events. It
     * does NOT remove it from any other list! The removeListener method has to be used
     * instead.
     */
    public void removeListenerAllEvents(EventListener listener){
        allEvents.remove(listener);
    }

    /**
     * This method removes the listener from the list of the given event.
     */
    public void removeListener(String eventName, EventListener listener){
        List list=(List) lists.get(eventName);
        if (list!=null){
            list.remove(listener);
        }
    }

    /** This method should not be used directly, but adjusted appropriately.*/
    public void informListeners(String eventName, EventObject event){
        List list=(List) lists.get(eventName);
        EventListener listener;
        Iterator i;
        if (list!=null && !list.isEmpty()){
            i=list.iterator();
            while (i.hasNext()){
                listener=(EventListener) i.next();

// !!!!TO DO: inform listener here: ----------------------------------------------------------------
//                e.g. listener.process(event);

            }
        }
        if (!allEvents.isEmpty()){
            i=allEvents.iterator();
            while (i.hasNext()){
                listener=(EventListener) i.next();

// !!!!TO DO: inform listener here: ----------------------------------------------------------------
//            e.g. listener.process(event);

            }
        }
    }
}