
//Project:      mosaic
//Version:      0.3
//Copyright:    Copyright (c) 1998
//Author:       Herbert Praehofer, Johannes Sametinger, Alois Stritzinger
//Affiliation:  Johannes Kepler University, Linz, Austria

// -------------------------------------------------------------------------------------------------
// Class BasicModel      ---------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------
// programmed by: hp
// last modified by: hp, 14.12.98
// documented by: hp, 7.4.99

package mosaic.sim.kernel;

import java.util.*;
import java.awt.*;
import mosaic.sim.kernel.variable.*;
import mosaic.sim.kernel.simevent.*;

/**
* Class implementing a root class for model implementation.
* <p>
* Extends AbstractModel.
* Implements the interface of TimeEventListener.
* Implements the interface of VariableChangeListener.
* <p>
* BasicModel is a basic implementation of a dynamic model with time and variable change
* events.
* <p>
* To extend this class and eventually implement your models based on it specialize the following methods:
* <ul>
* <li> <code>initialize</code> to intialize the model state variables.
* <li> <code>variableChange</code> to react to variable change events of Variables the ModelComponent listens to
* (this must be added as VariableChangeListener to the Variable).
* <li> <code>processSimEvent</code> to react to time events of Times the ModelComponent listens to
* (this must be added as TimeEventListener to the Timer).
* <li> <code>getDebugText</code> to print out special debug informations in traces.
* </ul>
* <p>
* @see sim.kernel.ModelComponent
* @see sim.kernel.Timer
* @see sim.kernel.BasicActiveModel
*/
public class BasicModel extends AbstractModelComponent
    implements TimeEventListener, VariableChangeListener  {


// Constructors ------------------------------------------------------------------------------------

  /** Default constructor */
  public BasicModel() {
      this ("model");
  }

  /**  Constructor with parent model and label. */
  public BasicModel(String label) {
    super(label);
  }


// -------------------------------------------------------------------------------------------------
// interface TimeEventListener ---------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

  /** Reacts to TimeEvents from eventComponents.
  * Specialize this methods to implement reactions to TimeEvents
  */
  public void processSimEvent (SimEvent evt) {
      // no implementation
  }

  /** variable to hold the SimEventComponents (Timers) this model listens to */
  Vector simEventComponents = new Vector ();

  /** Adds a SimEventComponent. Is called from addSimEventListener of SimEventComponent. */
  public void addSimEventComponent (SimEventComponent c){
    if (! simEventComponents.contains(c))
      simEventComponents.add(c);
  }

  /** Removes a SimEventComponent. Is called from removeSimEventListener of SimEventComponent.  */
  public void removeSimEventComponent (SimEventComponent c){
    if (! simEventComponents.contains(c))
      simEventComponents.add(c);
  }


   /** Adds a TimeEventComponent. Is called from addTimeEventListener of TimeEventComponent. */
   public void addTimeEventComponent (SimEventComponent c) {
      addSimEventComponent(c);
   }

   /** Removes a TimeEventComponent. Is called from removeTimeEventListener of TimeEventComponent.  */
   public void removeTimeEventComponent (SimEventComponent c){
      removeSimEventComponent(c);
   }


// -------------------------------------------------------------------------------------------------
// interface VariableChangeListener ----------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

  /** React to variable changed events. */
  public void variableChange (VariableChangeEvent e) {
      // no implementation
  }

  /** React to property changed event. */
  public void propertyChange (java.beans.PropertyChangeEvent e) {
    // no implementation
  }

  /** Add the variable to the list of variables this model listens to. */
  public void addListenerVariable(Variable v) {
      if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
        getSimApplication().getIntegrator().stateEventOccurred = true;
        return;
      }
      if (! listenerVariables.contains(v)) {
         listenerVariables.add(v);
      }
  }

  /** Remove the variable from the list of variables this model listens to. */
  public void removeListenerVariable (Variable v) {
    if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
      getSimApplication().getIntegrator().stateEventOccurred = true;
      return;
    }
    listenerVariables.remove(v);
  }

  /** vector of variables this model listens to. */
  protected Vector listenerVariables = new Vector();

  /** Gets the vector of variables this model listens to. */
  public Iterator getListenerVariables () {
    return listenerVariables.iterator();
  }

  /** Gets nth variables this model listens to. */
  public Variable getNthListenerVariable (int i) {
    if (i >= listenerVariables.size())
      return null;
    else
      return ((Variable) listenerVariables.get(i));
  }

// priority ----------------------------------------------------------------------------------------

    /** variable to hold the priority for handling this listener in handling variable change events. */
    protected int priority;

    /**
     * Gets the priority for the handling of variable change events by the simulator (the higher
     * the priority, the earlier it will be handled.)
     */
    public final int getPriority() {
        return priority;
    }

    /**
     * Sets the priority for the handling of variable change events by the simulator (the higher
     * the priority, the earlier it will be handled.)
     */
    public final void setPriority(int priority) {
        this.priority = priority;
    }


} // end BasicModel  -----------------------------------------------------------------------------------------


