/**
 * Project mosaic
 *
 * ModelComponent.java
 *
 * copyright 2000, JKU Linz, Austria
 */

package mosaic.sim.kernel;

import java.util.*;
import mosaic.sim.*; 

/**
*  Interface which all components in the mosaic simulation framework should implement.
*  <p>
*  The following features are specified in this interface definition for all mosaic simulation components.
*  <p>
*  (1) It defines methods to access the simulation application and simulation time.
*  <p>
*  (2) Model components are organized in a model component hierarchy. A model component becomes part
*  of a hierarchy, when it is added to a parent component already part in the hierarchy.
*  *  <p>
*  (3) Model components signal model components events when some significant changes occur for them,
*  e.g. when a component is added/removed as a child or when the component is enabled/disabled.
*  Those events are used by the simulator to cope with structural changes during a simulation run.
*/
public interface ModelComponent {

// constants to represent infinity and - infinity simulation time.

    /** constant representing infinity */
    public static final double INF = java.lang.Double.POSITIVE_INFINITY;
    /** constant representing infinity */
    public static final double INFINITY = java.lang.Double.POSITIVE_INFINITY;
    /** constant representing infinity */
    public static final double POSITIVE_INFINITY = java.lang.Double.POSITIVE_INFINITY;

    /** constant representing negative infinity */
    public static final double NEGINF = java.lang.Double.NEGATIVE_INFINITY;
    /** constant representing negative infinity */
    public static final double NEGINFINITY = java.lang.Double.NEGATIVE_INFINITY;
    /** constant representing negative infinity */
    public static final double NEGATIVE_INFINITY = java.lang.Double.NEGATIVE_INFINITY;

// getting the simulation application and current simulation time ----------------------------------

    /**
     * Gets the current simulation time for the simulation application this component belongs to.
     * @ return       the current simulation time as a double value
     */
    public double getTime();

    /**
     * Gets the simulation application object this component belongs to.
     * @return       the simulation application object
     */
    public SimApplication getSimApplication();

   /**
     * Sets the simulation application object this component belongs to.
     * @param application       the simulation application object
     */
    public void setSimApplication(SimApplication application);

// reset -------------------------------------------------------------------------------------------

    /** Resets the component to a predefined state. */
    public void reset ();

// name and info -----------------------------------------------------------------------------------

    /*
     * Gets the name of the component.
     * @return        the name of the component
     */
    public String getLabel();


    /**
     * Generates an information string for the component.
     * @return        the information string
     */
    String getInfo();


// maintaing the component hierarchy ---------------------------------------------------------------

    /**
     * Adds a component to this component.
     * @param component   the ModelComponent to be added to this component
     * @return true if addition was successful
     */
    public boolean addModelComponent(ModelComponent component);

    /**
     * Removes a component from this component.
     * @param component   the ModelComponent to be removed from this component
     * @return true if addition was successful
     */
    public boolean removeModelComponent(ModelComponent component);

    /**
     * Gets the parent ModelComponent.
     * @return        the ModelComponent which is the parent of this component
     */
    public ModelComponent getParentModelComponent ();

    /**
     * Sets the parent ModelComponent.
     * @param parent  the ModelComponent which becomes the parent of this component
     */
    public void setParentModelComponent (ModelComponent parent);

    /**
     * Gets a collection with all components.
     * @return        a collection containing all sub-component of this component
     */
    public Collection getModelComponents ();

    /**
     * Tests if the component is contained in this ModelComponent.
     * @return        true if the component is contained in this component
     */
    public boolean containsModelComponent (ModelComponent component);

// initialize --------------------------------------------------------------------------------------

    /** Method to initialize of the component. */
    public void initialize();

// enabled status   --------------------------------------------------------------------------------

    /** Tests if this component is enabled */
    public boolean isEnabled();

    /** Tests if this component is disabled */
    public boolean isDisabled();

    /** Sets the enabled status of this component */
    public void setEnabled(boolean s);

    /** Enable the model component */
    public void enable ();

    /** Disable the model component */
    public void disable ();

// ModelComponentEvent -----------------------------------------------------------------------------

    /**
    * Adds ModelComponentEventListener to this component.
    * @param listener   the listener to be added
    */
    public void addModelComponentEventListener(ModelComponentEventListener listener);

    /**
    * Removes ModelComponentEventListener from this component.
    * @param listener   the listener to be removed
    */
    public void removeModelComponentEventListener(ModelComponentEventListener listener);

    /**
    * Adds ModelComponentEventListener to this component for the event with name <code>eventName</code>. 
    * @param eventName  the name of the event
    * @param listener   the listener to be added
    */
    public void addModelComponentEventListener(String eventName, ModelComponentEventListener listener);

    /**
    * Removes ModelComponentEventListener to this component for the event with name <code>eventName</code>.
    * @param eventName  the name of the event
    * @param listener   the listener to be removed
    */
    public void removeModelComponentEventListener(String eventName, ModelComponentEventListener listener);


// fire model component events ---------------------------------------------------------------------

    /**
     * Fires an addition event for this model component
     * @param element the element which has been added
     */
    public void fireModelComponentAdded (ModelComponent element);

    /**
     * Fires a removal event for this model component.
     * @param element the element which has been removed
     */
    public void fireModelComponentRemoved (ModelComponent element);

    /**
     * Fires an enabled event for this model component.
     */
    public void fireModelComponentEnabled ();

    /**
     * Fires a disabled event for this model component.
     */
    public void fireModelComponentDisabled ();

    /**
     * Fires an event when this model component got continuous.
     */
    public void fireModelComponentGotContinuous();

    /**
     * Fires an event when this model component got discrete.
     */
    public void fireModelComponentGotDiscrete();

    /**
     * Fires an event for this model component when it has changed.
     */
    public void fireModelComponentChanged ();

    /**
     * Fires an event for this model component when it has changed.
     * @param e The object describing the event.
     */
    public void fireModelComponentChanged (ModelComponentEvent e);

} // end ModelComponent ============================================================================
