/**
 * Project mosaic
 *
 * AbstractModelComponent.java
 *
 * copyright 2000, JKU Linz, Austria
 */

package mosaic.sim.kernel.continuous;

import java.util.*;
import java.lang.*;
import java.lang.reflect.*;
import mosaic.sim.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.variable.*;
import mosaic.util.LogPrinter;
import mosaic.sim.kernel.StaticProjectSettings;
import mosaic.sim.kernel.simevent.*;

/**
* Class realizing the continuous simulation.
* It is responsible to set up and maintain the internal data structures used for the integration
* based on the messages on structural changes (adding and removal, enabling and disabling) of
* model elements (continuous variables, continuous equations, connections etc.  ).
* Realizes the integration step with the use of the integration method object. 
*/
public final class Integrator extends Object {

  public static boolean[] DEBUG = new boolean[] {true,
                                                false, // output to system.out
                                                false  //create LOG Files
                                                };

   /** Used as counter to divide signalVariablechanges by Constant sampleRate */
   protected static int sampleRate = 20;

   protected int counter = 0;


// Data structures ---------------------------------------------------------------------------------

   /** Boolean variable to test if the VariableChangeListeners are executed
   * to test for state events only. Used by the Variables' setValue methods. */
   public boolean isInIntegrationStep = false;

   /** Boolean variable to test if the VariableChangeListeners are executed
   * to test for state events only. Used by the Variables' setValue methods. */
   public boolean testForStateEvents = false;

   /** Boolean variable to signal if a state event occurred in the VariableChangeListeners.
   * Set in the Variables' setValue methods. */
   public boolean stateEventOccurred = false;

   /** ArrayList to hold the current list of VariableChangeListeners which have state events */
   ArrayList newVCLsWithEvent = new ArrayList();

   /** ArrayList to hold the list of VariableChangeListeners which have state events after the integration step */
   ArrayList vCLsWithEvent = new ArrayList();

   /** ArrayList to hold the current list of StateEventComponents which have state events */
   ArrayList newStateGuardsWithEvent = new ArrayList();

   /** ArrayList to hold the list of StateEventComponents which have state events after the integration step */
   ArrayList stateGuardsWithEvent = new ArrayList();


   /** Is used as flag, whether the model posessed by the integrator is valid or not. */
   protected boolean valid;


   protected boolean stateEventThrown;

   protected double normalStepWidth;

   /** Contains the stateEventTolerance value for the state event detection.*/
   protected double stateEventTolerance;

   /** Actual Simulation Time */
   protected double integrationTime;

   /** Holds the number of invalid Modelelements */
   protected int invalidCounter=0;

   /** ArrayList of ContEquations in system. */
   public ArrayList continuousEquations = new ArrayList();

   /**
   * Contains all ContEquations served by the intgerator in the order their
   * <code>equation()</code>-method has to be called.
   * check: used??
   */
   protected ArrayList sortedEquations = new ArrayList();

   /** Contains all variables of all models served by the integrator.*/
   protected ArrayList variables = new ArrayList();

   /** Contains all state variables of all models served by the integrator.*/
   protected ArrayList states = new ArrayList();

   /** Object doing the integration.  */
   protected IntegrationMethod integrationMethod;

   /** Contains all Stateguards for added variables. */
   protected ArrayList guards = new ArrayList();

   /** Contains variables which are known. */
   protected ArrayList knownVariables = new ArrayList();

   /** Contains all connections.*/
   protected ArrayList connections = new ArrayList();

   /** The EventRegistry for the checker objects. */
   protected final EventRegistry checkersReg=new EventRegistry();

   /** All EquationSorter in a hashlist */
   protected final Hashtable equationSorters=new Hashtable();

   /** All EquationSorters with empty domain */
   protected final ArrayList eqSortersNoDomain=new ArrayList(20);

   /** Equations have to be sorted again. */
   boolean resort=true;

   /** Number of enabled continuous equations */
   protected int enabledContEquations=0;

   /** ArrayList of all VariableListener of DoubleVariables */
   ArrayList contVariableListeners = new ArrayList();

   /** check: ArrayList of  counters of VariableListeners of DoubleVariables */
   ArrayList contVariableListenersCounter=new ArrayList();


// -------------------------------------------------------------------------------------------------
// constructors    ---------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

   public Integrator(SimApplication application) {
      this(application, new Heun(), 0.1);
   } // Integerator

   public Integrator(SimApplication application, double _stepWidth) {
      this(application, new Heun(), _stepWidth);
   } // Integrator

   public Integrator(SimApplication application, IntegrationMethod _integrationMethod) {
      this(application, _integrationMethod, 0.1);
   } // Integrator

   public Integrator(SimApplication application, IntegrationMethod _integrationMethod, double _stepWidth) {
      super();
      setSimApplication(application); 
      setIntegrationMethod(_integrationMethod);
      setStepWidth(_stepWidth);
      valid           = false;
      integrationTime = 0.0;
      stateEventTolerance = getStepWidth()/10;
   } // Integrator


// -------------------------------------------------------------------------------------------------
// access functions --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

// SimApplication ----------------------------------------------------------------------------------

    protected SimApplication application;

    /** Gets the simulation application object */
    public SimApplication getSimApplication() {
        return application;
    }

    /** Sets the simulation application object */
    public  void setSimApplication(SimApplication application) {
        this.application = application;
    }


// integration method ------------------------------------------------------------------------------

   /**
   * Sets the mathematical method which is used for integration.
   */
   public final void setIntegrationMethod(IntegrationMethod _integrationMethod) {
      this.integrationMethod = _integrationMethod;
      invalidate(); 
   } // setIntegrationMethod;

   /**
   * Sets the mathematical method which is used for integration from a text string.
   */
   public final void setIntegrationMethodAsText(String im) {
      String[] methods = StaticProjectSettings.getIntegrationMethods();
      boolean valid = false;
      if (im.equals(methods[0])) {
          setIntegrationMethod(new Euler());
          valid = true;
      } else if (im.equals(methods[1])) {
          setIntegrationMethod(new Heun());
          valid = true;
      } else if (im.equals(methods[2])) {
          setIntegrationMethod(new RungeKutta4());
          valid = true;
      } else {
          setIntegrationMethod(new Heun());
          valid = true;
          System.err.println(" *  Integration Method NOT RECOGNIZED set *");
      }
   }

   /**
   * Gets the mathematical method which is used for integration.
   */
   public final IntegrationMethod getIntegrationMethod() {
       return integrationMethod;
   }

   /**
   * Gets the mathematical method which is used for integration from a text string.
   */
   public final String getIntegrationMethodAsText() {
      String[] methods = StaticProjectSettings.getIntegrationMethods();
      String ret;
      if (integrationMethod instanceof Euler) {
         ret = methods[0];
      } else if (integrationMethod instanceof Heun) {
         ret = methods[1];
      } else if (integrationMethod instanceof RungeKutta4) {
         ret = methods[2];
      } else {
         ret = null;
         System.err.println(" *  Integration method NOT RECOGNIZED in get*");
      }
      return ret;
   }


// step width --------------------------------------------------------------------------------------

   /**
   * Returns the stepwidth for the next integration step
   */
   public final double getStepWidth()  {
      return normalStepWidth;
   }

   /**
   * Sets the stepwidth for the next integration step.
   */
   public final void setStepWidth(double _stepWidth) {
       this.normalStepWidth = _stepWidth;
   }


// state event tolerance ---------------------------------------------------------------------------

   /**
   * Returns the stateEventTolerance value for state event detection.
   */
   public final double getStateEventTolerance() {
      return stateEventTolerance;
   }

   /**
   * Sets the  stateEventTolerance value for state event detection.
   */
   public final void setStateEventTolerance(double _stateEventTolerance) {
      this.stateEventTolerance = _stateEventTolerance;
   }


// sample rate -------------------------------------------------------------------------------------
// check: to be deleted

   /**
   * Returns the rate for the output of continuous values
   */
   public final int getSampleRate()  {
      return sampleRate;
   }

   /**
   * Sets the sample rate  for the next integration step.
   */
   public final void setSampleRate(int _rate) {
       this.sampleRate = _rate;
   }


// -------------------------------------------------------------------------------------------------
// maintaining the structure -----------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

// addition and removal of components --------------------------------------------------------------

   /**
   * Adds the continousEqation, if not already added, to the integrator.
   */
   public final void add(ContEquation ce)  {
      if (!continuousEquations.contains(ce))
         continuousEquations.add(ce);
      equationSorters.put(ce,new EquationSorter(ce));
      new EquationChecker(ce);
   }

   /**
   * Adds a continuous (!) variable, if not already added, to the integrator.
   */
   public final void add(DoubleVariable v)  {

      if (!variables.contains(v))
         variables.add(v);

      if ((v.isTypeState()) && (!states.contains(v)))
         states.add(v);
      new VariableChecker(v);
      Collection listeners=v.getVariableChangeListener();
      if (listeners!=null && listeners.size()>0){
        Iterator i=listeners.iterator();
        while (i.hasNext()){
          addDoubleVariableListener((VariableChangeListener)i.next());
        }
      }
   } // add

   /**
   * Adds connection, if not already added, to the integrator.
   */
   public final void add(Connection c){
     if (connections.contains(c))return;
     connections.add(c);
     new ConnectionChecker(c);
   }

   /**
   * Adds the stateguard, if not already added, to the integrator.
   */
   public final void add(StateEventComponent stateGuard) {
      if (!guards.contains(stateGuard))
         guards.add(stateGuard);
      invalidate();
   }

   /**
   * Removes the stateguard, if exists in the guardlist, from the integrator.
   */
   public final void remove(StateEventComponent stateGuard) {
      if (guards.contains(stateGuard))
         guards.remove(stateGuard);
      invalidate();
   } // remove

   /**
   * Removes the connection from the integrator.
   */
   public final void remove(Connection c){
     connections.remove(c);
     checkersReg.removal(c);
   }

   /**
   * Removes the equation from the integrator.
   */
   public final void remove(ContEquation ce){
     continuousEquations.remove(ce);
     checkersReg.removal(ce);
     equationSorters.remove(ce);
   }

   /**
   * Removes the Variable from the interator.
   */
   public final void remove(Variable v){
     variables.remove(v);
     states.remove(v);
     checkersReg.removal(v);
   }


   public void addDoubleVariableListener(VariableChangeListener l) {
        int pos=contVariableListeners.indexOf(l);
        if (pos<0) {
            contVariableListeners.add(l);
            contVariableListenersCounter.add(new Integer(1));
        }else{
            Integer num=(Integer)(Integer)contVariableListenersCounter.get(pos);
            contVariableListenersCounter.set(pos,new Integer(num.intValue()+1));
        }
    }

    public void removeDoubleVariableListener(VariableChangeListener l) {
        int pos=contVariableListeners.indexOf(l);
        Integer num=(Integer)contVariableListenersCounter.get(pos);
        if(num.intValue()<=1){
            contVariableListenersCounter.remove(pos);
            contVariableListeners.remove(pos);
        }else{
            contVariableListenersCounter.set(pos,new Integer(num.intValue()+1));
        }
    }


// enabling and disabling --------------------------------------------------------------------------

   /**
   * Tells the integrator the enabling of the connection.
   */
   public final void enable(Connection c){
     checkersReg.enabling(c);
   }

   /**
   * Tells the integrator the enabling of the equation.
   */
   public final void enable(ContEquation ce){
     checkersReg.enabling(ce);
   }


   /**
   * Tells the integrator the enabling of the variable.
   */
   public final void enable(Variable v){
     checkersReg.enabling(v);
   }

   /**
   * Tells the integrator the disabling of the connection.
   */
   public final void disable(Connection c){
     checkersReg.disabling(c);
   }

   /**
   * Tells the integrator the disabling of the equation.
   */
   public final void disable(ContEquation ce){
     checkersReg.disabling(ce);
   }


   /**
   * Tells the integrator the disabling of the variable.
   */
   public final void disable(Variable v){
     checkersReg.disabling(v);
   }


// change of components ----------------------------------------------------------------------------

   /**
   * Tells the integrator the change of the connection.
   */
   public final void change(Connection c){
     checkersReg.change(c);
   }

   /**
   * Tells the integrator the change of the equation.
   */
   public final void change(ContEquation ce){
     checkersReg.change(ce);
   }

   /**
   * Tells the integrator the change of the variable.
   */
   public final void change(Variable v){
     checkersReg.change(v);
     if (v.isTypeState()) if (!states.contains(v)) states.add(v);
     else if (states.contains(v)) states.remove(v);
   }

   /**
   * Used by the checkers. If an element gets valid, the counter is decreased by 1.
   */
   public final void decInvalid(){
     invalidCounter--;
   }

   /**
   * Used by the checkers. If an element gets invalid, the counter is increased by 1.
   */
   public final void incInvalid(){
     invalidCounter++;
   }


// -------------------------------------------------------------------------------------------------
// equation sorting --------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

   /** Finds all equations that are connected (directly or indirectly) to the given, and
 *    have to be executed before.*/
   protected final static ArrayList searchLeft(ContEquation c){
     Variable d[]=c.getDomain();
     ArrayList result=new ArrayList(10);
     if (c==null) return result;
     for (int i=0;i<d.length;i++){
       ContEquation ce=searchLeft(d[i]);
       if (ce!=null) result.add(ce);
     }
     return result;
   }

   /** Finds the equation that is connected (directly or indirectly) to the given variable, and
 *    has to be executed before.*/
   protected final static ContEquation searchLeft(Variable v){
     if (v==null) return null;
     if (v.getDefinedBy()!=v)v=v.getDefinedBy();
     if (!(v.isTypeState()||v.isTypeConstant()||v.isDisabled())){
       Coupler c = v.getInCoupler();
       if (c!=null && c instanceof ContEquation){
         return (ContEquation) c;
       }
     }
     return null;
   }

   /** Finds all equations that are connected (directly or indirectly) to the given, and
 *    have to be executed afterwards.*/
   protected final static ArrayList searchRight(ContEquation c){
     Variable targets[]=c.getTarget();
     ArrayList result=new ArrayList();
     if (c==null||targets==null) return result;
     for (int i=0;i<targets.length;i++){
       result.addAll(searchRight(targets[i]));
     }
     return result;
   }

   /** Finds the equation that is connected (directly or indirectly) to the given variable, and
 *    has to be executed afterwards.*/
   protected final static ArrayList searchRight(Variable v){
     ArrayList result=new ArrayList();
     if (v==null || v.isTypeState() || v.isTypeConstant()||v.isDisabled()) return result;
     List targest=v.getOutCouplers();
     Iterator enum=targest.iterator();
     while (enum.hasNext()){
       Object o=enum.next();
       if (o!=null){
         if (o instanceof ContEquation) result.add(o);
         else if (o instanceof Connection){
           Connection c=(Connection) o;
           result.addAll(searchRight(c.getTarget()));
         }
       }
     }
     return result;
   }

   /** Sort again. */
   protected final void resort(){
     resort=true;
     invalidate();
   }


   /**
   * Sets the state of the integrator to invalid and causes the integrator
   * to build up the internal structures again.
   */
   public final void invalidate() {
      valid = false;
   } // invalidate

   /**
   * Checks if equations or stateVariables exist
   */
   public boolean continuousExist() {
     return (!(this.continuousEquations.isEmpty() && this.states.isEmpty())) ;
   }

// sortEquations -----------------------------------------------------------------------------------

   /**
   * Determines the correct sort order of the models for the execution of
   * the <code>equation</code>-method.
   */
   public final void sortEquations()
      throws ContSimulationException {
      if (invalidCounter!=0){
         String message="Modelelements invalid: \n";
         Enumeration enum=checkersReg.hash.elements();
         while (enum.hasMoreElements()){
           Iterator enum2=((ArrayList)enum.nextElement()).iterator();
           while (enum2.hasNext()){
             ElementChecker ch=(ElementChecker)enum2.next();
             if (!ch.getValidStatus()){
               message=message+ch.who()+"\n";
             }
           }
         }
         throw new CSESortingError(message);
      }
      if (resort){
        sortedEquations.clear();
        Iterator start=eqSortersNoDomain.iterator();
        EquationSorter sorter;
        while (start.hasNext()){
          sorter=(EquationSorter)start.next();
          sorter.sortedDomain();
        }
        if (sortedEquations.size()!=enabledContEquations) throw new CSESortingError();
        resort=false;
      }
      valid=true;

  } // sortEquations


// -------------------------------------------------------------------------------------------------
// integrationStep ---------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

   /**
   * Executes the next step in the simulation.
   */
   public final boolean integrationStep(double untilTime, double stepWidth)
      throws ContSimulationException {

      counter++;
      if (counter == sampleRate)
            counter = 0;
      isInIntegrationStep = true;
      stateEventThrown = false;
      double actStepWidth = stepWidth;

      // eventually initialize  integrationMethod
      if (integrationMethod == null) {
         throw(new CSENoIntegrationMethod());
      } else {
         if (!valid) {
            // sort the equations new and intialize the
            sortEquations();
            integrationMethod.initialize(states, sortedEquations);
            integrationMethod.doEquations();
            // forward initial values to views
            signalVariableChanges();  // check: if valid ( what is with fireVariableChanges)
         }
      }

      // save current state in previous
      for (int i = 0; i < states.size(); i++) {
        DoubleVariable state = ((DoubleVariable) states.get(i));
        state.saveToPrevious();
      }

      // make the integration step
      integrationMethod.integrationStep(integrationTime, actStepWidth);

      // check for state events  -----------------------------------------------
      stateGuardsWithEvent = checkStateGuards(guards);
      vCLsWithEvent = checkVariableChangeListeners(contVariableListeners);
      if (! stateGuardsWithEvent.isEmpty() || ! vCLsWithEvent.isEmpty()) {
        // iterate to state events
        double lowerBound = 0.0;
        double upperBound = actStepWidth;

        boolean ok = false;
        // while not iterated exactly do
        while(!ok) {
            // compute the actStepWidth to lie in the middle of lower and upperBound
            if (upperBound - lowerBound <= stateEventTolerance) {
                actStepWidth = upperBound;
                ok = true;
            }
            else
                actStepWidth = lowerBound + (upperBound - lowerBound) / 2.0;

            // restore the states to previous states
            Iterator stateIterator = states.iterator();
            while(stateIterator.hasNext()) {
                DoubleVariable state = (DoubleVariable)stateIterator.next();
                state.setValue(state.getPreviousValue());
            }

            // make integration step with new step width
            integrationMethod.integrationStep(integrationTime, actStepWidth);

            newStateGuardsWithEvent = checkStateGuards(stateGuardsWithEvent);
            newVCLsWithEvent = checkVariableChangeListeners(vCLsWithEvent);
            // check for events and compute new stepWidth
            if (! newStateGuardsWithEvent.isEmpty() || ! newVCLsWithEvent.isEmpty()) {
                // state event at actStepWidth detected
                stateGuardsWithEvent = newStateGuardsWithEvent;
                vCLsWithEvent = newVCLsWithEvent;
                if ((upperBound - actStepWidth) <= stateEventTolerance)
                     ok = true;
                upperBound = actStepWidth;
            }
            // no state event at actStepWidth detected
            else {
                lowerBound = actStepWidth;
                //upperBound unchanged
            }
        }

        // set integration time to old integration time + actual step width
        integrationTime = integrationTime + actStepWidth;

      } // end check for state events --------------------------------------------

      // make normal step without state events and set inegration time to untilTime
      else {
        integrationTime = untilTime;
      } // end if then else

      // set simulation time to integration time
      getSimApplication().setTime(integrationTime);

      // do the equations to set all variables.
      integrationMethod.doEquations();

      // end of the integration step - set flag false 
      isInIntegrationStep = false;

      // execute state events
      if (! stateGuardsWithEvent.isEmpty()) {
        throwStateGuards(stateGuardsWithEvent);
        stateEventThrown = true;
      }
      if (! vCLsWithEvent.isEmpty()) {
        stateEventThrown = true;
      }

      // signal the variable changes to all listeners of continuous variables.
      if (stateEventThrown || counter == 0)
        signalVariableChanges();

      // return the time until integration occurred
      return stateEventThrown;

   } // end integrationStep  -----------------------------------------------------------------------


   /**
   * Checks the list of stateguards, if any event occured.
   */
   protected ArrayList checkStateGuards(ArrayList stateGuards) {
      ArrayList nextStateGuards = new ArrayList();

      for(int loop = 0;loop < stateGuards.size();loop++) {
         if (((StateEventComponent)(stateGuards.get(loop))).eventOccured()) {
            nextStateGuards.add(stateGuards.get(loop));
         }
      }
      return nextStateGuards;
   } // checkStateGuards

   /**
   * Checks the list of stateguards, if any event occured.
   */
   protected ArrayList checkVariableChangeListeners(ArrayList vCListeners) {
      ArrayList nextVCListeners = new ArrayList();

      VariableChangeListener listener;
      testForStateEvents = true;
      for(int loop = 0;loop < vCListeners.size();loop++) {
         stateEventOccurred = false;
         listener = ((VariableChangeListener)(vCListeners.get(loop)));
         listener.variableChange(null);
         if (stateEventOccurred) {
            nextVCListeners.add(listener);
         }
      }
      testForStateEvents = false;
      return nextVCListeners;
   } // checkVariableChangeListeners


   /**
   * Throws all events that occured in the list of stateguards.
   */
   protected void throwStateGuards (ArrayList stateGuards) {
      for(int loop=0;loop < stateGuards.size();loop++)
         if (((StateEventComponent)(stateGuards.get(loop))).eventOccured())
            ((StateEventComponent)(stateGuards.get(loop))).triggerEvent();
   } // throwStateGuards



    /** Signal the changes of the continuous variables. */
    public final void  signalVariableChanges() {
        Iterator varElems = variables.iterator();
        Variable variable;
        while(varElems.hasNext()) {
            variable = (Variable)varElems.next();
            if (variable instanceof DoubleVariable && ((DoubleVariable)variable).isContinuous()) {
                ((DoubleVariable)variable).signalVariableChangeDirectly();
            }
        }
    }


// -----------------------------------------------------------------------------
// maintaining
// -----------------------------------------------------------------------------

// error and debug output --------------------------------------------------------------------------

    protected final void errorOutputAlgebraicLoop (ArrayList equations) {
        Iterator equationElems = equations.iterator();
        ContEquation equation;
        getSimApplication().errorOutput("integration", "Algebraic loop detected ---------------------------------------------------");

        while (equationElems.hasNext()) {
          equation = ((ContEquation)equationElems.next());

          Variable[] domain = equation.getDomain();
          getSimApplication().debugOutput("integration", "Domain of Equation:", 2);
          for (int i = 0; i < domain.length; i++) {
            if (domain[i].getParentModelComponent() != null) {
                getSimApplication().debugOutput("integration", "    Variable " + domain[i].getLabel()+
                                   " of model " + domain[i].getParentModelComponent().getLabel() +
                                   " involved.", 2);
                if ( !knownVariables.contains(domain[i]) ) {
                    getSimApplication().debugOutput("integration", " -->>> not known !!", 2);
                }
            } else {
                getSimApplication().debugOutput("integration", "    Variable " + domain[i].getLabel()+
                                   " involved.", 2);
                if ( !knownVariables.contains(domain[i]) ) {
                    getSimApplication().debugOutput("integration", " -->>> not known !!", 2);
                }
            } // end if then else
          } // end for

          Variable[] target = equation.getTarget();
          getSimApplication().debugOutput("integration", "Target of Equation:", 2);
          for (int i = 0; i < target.length; i++) {
            if (target[i].getParentModelComponent() != null) {
                getSimApplication().debugOutput("integration", "    Variable " + target[i].getLabel()+
                                       " of model " + target[i].getParentModelComponent().getLabel() +
                                       " involved.", 2);
                if ( !knownVariables.contains(target[i]) ) {
                        getSimApplication().debugOutput("integration", " -->>> not known !!", 2);
                }
            } else {
                getSimApplication().debugOutput("integration", "    Variable " + target[i].getLabel()+
                                      " involved.", 2);
                if ( !knownVariables.contains(domain[i]) ) {
                        getSimApplication().debugOutput("integration", " -->>> not known !!", 2);
                }
            } // end if then else
          } // end for
        } // end while
      } // end errorOutputAlgebraicLoop

// debug output by af ------------------------------------------------------------------------------

  protected final void printUnknownVariables(ArrayList equations) {
    Iterator equationElems = equations.iterator();
    ContEquation equation;
    if (DEBUG[1]) System.out.println("Algebraic loop detected ---------------------------------------------------");
    if (DEBUG[2]) LogPrinter.println("Algebraic loop detected ---------------------------------------------------");

    while (equationElems.hasNext()) {
      equation = ((ContEquation)equationElems.next());
      Variable[] domain = equation.getDomain();
      if (DEBUG[1]) {
          System.out.println("-------------------------------------------------------------------------");
          System.out.println("Domain of Equation:");
      }

      if (DEBUG[2]) {
          LogPrinter.println("-------------------------------------------------------------------------");
          LogPrinter.println("Domain of Equation:");
      }

      for (int i = 0; i < domain.length; i++) {
          if (DEBUG[1])
            if (domain[i].getParentModelComponent() != null) {
                System.out.println("    Variable " + domain[i].getLabel()+
                                   " of model " + domain[i].getParentModelComponent().getLabel() +
                                   " involved.");
                if ( !knownVariables.contains(domain[i]) )
                    System.out.println(" -->>> not known !!");
            }
            else {
                System.out.println("    Variable " + domain[i].getLabel()+
                                   //" of model " + domain[i].getParentModelComponent().getLabel() +
                                   " involved.");
                 if ( !knownVariables.contains(domain[i]) )
                    System.out.println(" -->>> not known !!");
            }

          if (DEBUG[2]) LogPrinter.println("    Variable " + domain[i].getLabel()+
                                           //" of model " + domain[i].getParentModelComponent().getLabel() +
                                           " involved.");

      }
      Variable[] target = equation.getTarget();
      if (DEBUG[1]) System.out.println("Target of Equation:");
      if (DEBUG[2]) LogPrinter.println("Target of Equation:");
      for (int i = 0; i < target.length; i++) {
          if (DEBUG[1]) {
            if (target[i].getParentModelComponent() != null)
                System.out.println("    Variable " + target[i].getLabel()+
                                   " of model " + target[i].getParentModelComponent().getLabel() +
                                   " involved.");
                 if ( !knownVariables.contains(domain[i]) )
                    System.out.println(" -->>> not known !!");
            }
            else {
                System.out.println("    Variable " + target[i].getLabel()+
                                   //" of model " + target[i].getParentModelComponent().getLabel() +
                                   " involved.");
                if ( !knownVariables.contains(domain[i]) )
                    System.out.println(" -->>> not known !!");
            }
          if (DEBUG[2]) LogPrinter.println("    Variable " + target[i].getLabel()+
                                           //" of model " + target[i].getParentModelComponent().getLabel() +
                                           " involved.");
      }


    }

  }



//--------------------------------------------------------------------------------------------------
// Inner classes -----------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------

   /** The ElementChecker checks whether a given ModelComponent is valid or not.
    * If it is not valid, the invalidCounter is increased by one. If the element
    * changes it's state than the invalidCounter is updated by this class. */
   abstract class ElementChecker {
     final ModelComponent elem; // the checked element
     boolean prevValid=true; // the last known state

     public ElementChecker(ModelComponent e){
       elem=e;

       // register at EventRegistry;
       checkersReg.add(this,e);
       update(); // update state
     }

     /** Tells whether the element is currently valid. */
     public final boolean getValidStatus(){return prevValid||elem.isDisabled();}

     /** recheck */
     public final void update(){
       if (elem.isDisabled()){ // if checked element is disabled then remember new state only.
         prevValid=check();
       }
       else{ // if enabled than tell the integrator the new state.
         String l=elem.getLabel();
         boolean valid=check();
         if (valid!=prevValid){ // if the state stays the same do nothing.
           if (prevValid) incInvalid(); //got invalid
           else decInvalid(); // got valid
           prevValid=valid;
         }
       }
     }

     /** The real checking is done in this method. */
     public abstract boolean check();

     /** if element gets removed than remove this as listener from parent. */
     public void modelElementRemoved (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       checkersReg.remove(this,e);
     }

     /** handle the enabeling */
     public void modelElementEnabled (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       if (!prevValid) incInvalid();
       update();
     }

     /** handle the disabling */
     public void modelElementDisabled (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       if (!prevValid) decInvalid();
     }

     /** update state if element changed */
     public void modelElementChanged (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       update();
     }

     /** Returns the label of the checked element. Used for the error message. */
     public String who(){
       return elem.getLabel();
     }
  } // end ElementChecker --------------------------------------------------------------------------

   /** Special checker for equations. */
   final class EquationChecker extends ElementChecker{

     final EquationSorter sorter;

     public EquationChecker(ContEquation ce){
       super(ce);
       sorter=(EquationSorter)equationSorters.get(ce);
       if (ce.isEnabled())enabledContEquations++;
     }

     public final boolean check(){
       int i;
       ContEquation ce=(ContEquation)elem;

       // check all target variables:
       //   no constants!
       //   no states!
       //   no disabled (if equation is enabled)!
       Variable targets[]=ce.getTarget();
       Variable target;
       for (i=0;i<targets.length;i++){
         target=targets[i];
         if (target.isTypeConstant()||target.isTypeState()||target.isDisabled()) return false;
       }
       // check all domain variables:
       //   no disabled if equation is enabled!
       Variable domain[]=ce.getDomain();
       for (i=0;i<domain.length;i++){
         if (domain[i].isDisabled()) return false;
       }
       return true;
     }

     public final void modelElementRemoved(ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.

//       EquationSorter sorter=(EquationSorter) equationSorters.get(e);
       sorter.disable();
       enabledContEquations--;

       checkersReg.remove(this,e);
     }

     // handle the enabeling
     public void modelElementEnabled (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       sorter.enable();
       if (!prevValid) incInvalid();
       update();
       enabledContEquations++;
     }

     // handle the disabling
     public void modelElementDisabled (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       sorter.disable();
       if (!prevValid) decInvalid();
       enabledContEquations--;
     }

     // update state if element changed
     public void modelElementChanged (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       sorter.disable();
       sorter.enable();
       update();
     }
   } // end EquationChecker ------------------------------------------------------------------------


   /** Special checker for connections. */
   final class ConnectionChecker extends ElementChecker{
      Variable prevTarget;Variable prevDomain;

     public ConnectionChecker(Connection c){
       super(c);
       prevTarget=c.getTarget();
       prevDomain=c.getDomain();
     }

     public final boolean check(){
       Connection c=(Connection)elem;
       Variable prevDomain;
       Variable prevTarget;

       // check target variable:
       //   not constant!
       //   not state!
       //   not disabled if connection is enabled!

       Variable target=c.getTarget();
       if (target==null||target.isTypeConstant()||target.isTypeState()||target.isDisabled()) return false;

       // check domain variable:
       //   not disabled if connection is enabled!
       Variable domain=c.getDomain();
       if (domain==null||domain.isDisabled()) return false;
       return true;
     }

     /** This method removes connections from the partial order of the equations. */
     protected final void removeSorterConnection(ContEquation left, ArrayList rights){
       if (left==null||rights.isEmpty()||rights==null) return;
       Iterator enum=rights.iterator();
       ContEquation right;
       EquationSorter leftSorter,rightSorter;
       leftSorter=(EquationSorter)equationSorters.get(left);
       if (leftSorter!=null){
         while(enum.hasNext()){
           right=(ContEquation)enum.next();
           rightSorter=(EquationSorter)equationSorters.get(right);
           if (rightSorter!=null){
             leftSorter.removeTarget(rightSorter);
             rightSorter.removeDomain(leftSorter);
           }
         }
       }
     }

     /** This method establishes connections in the partial order of the equations. */
     protected final void createSorterConnection(ContEquation left, ArrayList rights){
       if (left==null||rights.isEmpty()||rights==null) return;
       Iterator enum=rights.iterator();
       ContEquation right;
       EquationSorter leftSorter,rightSorter;
       leftSorter=(EquationSorter)equationSorters.get(left);
       if (leftSorter!=null){
         while(enum.hasNext()){
           right=(ContEquation)enum.next();
           rightSorter=(EquationSorter)equationSorters.get(right);
           if (rightSorter!=null){
             leftSorter.addTarget(rightSorter);
             rightSorter.addDomain(leftSorter);
           }
         }
       }
     }

     // if element gets removed than remove this as listener from parent.
     public void modelElementRemoved (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       removeSorterConnection(searchLeft(prevDomain),searchRight(prevTarget));
       checkersReg.remove(this,e);
     }

     // handle the enabeling
     public void modelElementEnabled (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       if (!prevValid) incInvalid();
       createSorterConnection(searchLeft(prevDomain),searchRight(prevTarget));
       update();
     }

     // handle the disabling
     public void modelElementDisabled (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       removeSorterConnection(searchLeft(prevDomain),searchRight(prevTarget));
       if (!prevValid) decInvalid();
     }

     // update state if element changed
     public void modelElementChanged (ModelComponent e){
       if (!(e==elem)) return; // if it's the wrong one then do nothing.
       // possilbe speedup: calculate differences
       if (e.isEnabled()){
         removeSorterConnection(searchLeft(prevDomain),searchRight(prevTarget));
         prevDomain=((Connection)elem).getDomain();
         prevTarget=((Connection)elem).getTarget();
         createSorterConnection(searchLeft(prevDomain),searchRight(prevTarget));
       }
       else{
         prevDomain=((Connection)elem).getDomain();
         prevTarget=((Connection)elem).getTarget();
       }
       update();
     }
   } // end ConnectionChecker ----------------------------------------------------------------------


   final class VariableChecker extends ElementChecker{

     public VariableChecker(DoubleVariable v){
       super(v);
     }

     public final boolean check(){
       DoubleVariable v=(DoubleVariable)elem;
       // State:
       //   Derivative defined
       //   not constant
       if (v.isTypeState()){
         if (v.getDerivative()==null||v.isTypeConstant()) return false;
         return true;
       }
       // not State:
       //   Derivative = null
       if (!v.isTypeState()){
         if (v.getDerivative()!=null) return false;
       }
       return true;
     }

   } // end VariableChecker ------------------------------------------------------------------------


   /** Class EventRegistry holds all checkers and associates them with their ModelComponents. */
   final class EventRegistry{
     Hashtable hash=new Hashtable();
     public final void add(ElementChecker c, ModelComponent e){
       if (hash.containsKey(e)){
         // just add the checker to the list, if the element is already in the list.
         ArrayList checkers=(ArrayList)hash.get(e);
         if (checkers.contains(c)) return;
         checkers.add(c);
       }
       else{
         // add new entry
         ArrayList checkers=new ArrayList(2);
         checkers.add(c);
         hash.put(e,checkers);
       }
     }
     public final void remove(ElementChecker c, ModelComponent e){
       if (!hash.containsKey(e)) return;
       // remove the checker from the list.
       ArrayList checkers=(ArrayList)hash.get(e);
       checkers.remove(c);
       // if the list is empty, remove the key
       if (checkers.isEmpty()) hash.remove(e);
     }

     /** Signal removal to the checkers. The checker should remove himself.*/
     public final void removal(ModelComponent e){
       if (!hash.containsKey(e)) return;
       ArrayList checkers=(ArrayList)hash.get(e);
       Iterator enum=(new ArrayList(checkers)).iterator();
       while(enum.hasNext()){
         ElementChecker checker=(ElementChecker)enum.next();
         checker.modelElementRemoved(e);
       }
     }

     /** Signal enabling of the element to the checkers. */
     public final void enabling(ModelComponent e){
       if (!hash.containsKey(e)) return;
       ArrayList checkers=(ArrayList)hash.get(e);
       Iterator enum=(new ArrayList(checkers)).iterator();
       while(enum.hasNext()){
         ElementChecker checker=(ElementChecker)enum.next();
         checker.modelElementEnabled(e);
       }
     }

     /** Signal disabling of the element to the checkers. */
     public final void disabling(ModelComponent e){
       if (!hash.containsKey(e)) return;
       ArrayList checkers=(ArrayList)hash.get(e);
       Iterator enum=(new ArrayList(checkers)).iterator();
       while(enum.hasNext()){
         ElementChecker checker=(ElementChecker)enum.next();
         checker.modelElementDisabled(e);
       }
     }

     /** Signal change of the element to the checkers. */
     public final void change(ModelComponent e){
       if (!hash.containsKey(e)) return;
       ArrayList checkers=(ArrayList)hash.get(e);
       Iterator enum=(new ArrayList(checkers)).iterator();
       while(enum.hasNext()){
         ElementChecker checker=(ElementChecker)enum.next();
         checker.modelElementChanged(e);
       }
     }
   } //  end EventRegistry -------------------------------------------------------------------------


   /** This class is the representation of an equation in the partial order.
   * It is used to sort the equations.
   */
   final class EquationSorter{
     final ContEquation myEquation;
     private final ArrayList targets=new ArrayList(10);
     final ArrayList domains=new ArrayList(10);
     int sortedDomains=0;
     int domainSize=0;

     public EquationSorter(ContEquation ce){
       myEquation=ce;
       if (ce.isEnabled()) enable();
     }

     public void removeTarget(EquationSorter es){
       targets.remove(es);
     }

     public void addTarget(EquationSorter es){
       if (targets.contains(es)) return;
       targets.add(es);
     }

/*     public ArrayList getTargets(){
       return targets;
     }*/

     public void removeDomain(EquationSorter es){
       domains.remove(es);
       domainSize--;
       if (domainSize==0) eqSortersNoDomain.add(this);
     }

     public void addDomain(EquationSorter es){
       if (domains.contains(es)) return;
       if (domainSize==0) eqSortersNoDomain.remove(this);
       domainSize++;
       domains.add(es);
     }

     public ArrayList getDomains(){
       return domains;
     }

     /** This method tells the sorter, that one of its domains got sorted in.
      * If all domains got sorted in, then this one is sorted in and the targets
      * are informed.
     */
     public void sortedDomain(){
       sortedDomains++;
       if (sortedDomains>=domainSize){// all domains are sorted in.
         sortedEquations.add(myEquation);

         Iterator enum=targets.iterator();
         while (enum.hasNext()){
           EquationSorter es=(EquationSorter) enum.next();
           es.sortedDomain();
         }
         sortedDomains=0;
       }
     }

     /** establish connections */
     public void enable(){
       // Search left
       Iterator enum=searchLeft(myEquation).iterator();
       while (enum.hasNext()){
         EquationSorter es=(EquationSorter)equationSorters.get(enum.next());
         if (es!=null){
           addDomain(es);
           es.addTarget(this);
         }
       }

       //Search right
       enum=searchRight(myEquation).iterator();
       while (enum.hasNext()){
         EquationSorter es=(EquationSorter)equationSorters.get(enum.next());
         if (es!=null){
           addTarget(es);
           es.addDomain(this);
         }
       }
       if (domains.size()==0) eqSortersNoDomain.add(this);
       sortedDomains=0;
     }

     /* remove connections */
     public void disable(){
       Iterator enum=targets.iterator();
       EquationSorter target;
       while (enum.hasNext()){
         target=(EquationSorter)enum.next();
         target.removeDomain(this);
       }
       enum=domains.iterator();
       EquationSorter domain;
       while (enum.hasNext()){
         domain=(EquationSorter)enum.next();
         domain.removeTarget(this);
       }
       targets.clear();
       if (domainSize>0){
         domains.clear();
         domainSize=0;
       }
       else eqSortersNoDomain.remove(this);
     }
   } // end  EquationSorter ------------------------------------------------------------------------


} // class Integrator

// end of Integrator.java
