
//Project:      SimBeans kernel
//Version:      0.3
//Copyright:    Copyright (c) 1998
//Author:       Herbert Praehofer, Johannes Sametinger, Alois Stritzinger
//Affiliation:  Johannes Kepler University, Linz, Austria

// -----------------------------------------------------------------------------------------------------------
// interface BasicStatistic    ----------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------

package mosaic.sim.kernel.statistics;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.beans.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.control.*;

/**
* Base class for the statistics beans to calculate statistics of number values.
* Computes max and min value received.
* <p>
* Use this class for computing min and max of numeric properties
* <p>
*  <b>Bound properties:</b>
* <dl>
* <dt><code> min      </code></dt> <dd> the minimum value received </dd>
* <dt><code> max      </code></dt> <dd> the maximum value received </dd>
* </dl>
*
* @see   TimeDependentStatistic
* @see   ObservationalStatistic
*
* @author Herbert Praehofer
* @version %I% %G%
*/
abstract public class BasicStatistic extends AbstractModelComponent implements Statistic{

  /** Default constructor VariableStatistic */
  public BasicStatistic() {

  }

// -----------------------------------------------------------------------------------------------------------
// interface Statistic
// -----------------------------------------------------------------------------------------------------------


    /** report the statistic value to an output stream */
    public void report (java.io.DataOutputStream out) throws java.io.IOException  {
        out.writeBytes("  from start time " + startTime + " to current time " + getTime() + "\n\n");
        out.writeBytes("  Minimum:      " + getMin() + "\n");
        out.writeBytes("  Maximum:      " + getMax() + "\n");
        out.writeBytes("  Mean:         " + getMean() + "\n");
        out.writeBytes("  StdDeviation: " + getStandardDeviation() + "\n");
    }


// interface statistic
    /** Accept new value for the statistic */
    public void newObservation (Object newVal) {
        if (newVal instanceof Number) {
            setNumber ((Number)newVal);
            computeStatistics();
        }
        else if (newVal instanceof Boolean) {
            if (((Boolean)newVal).booleanValue() == true)
              setNumber (new Integer (1));
            else
              setNumber (new Integer (0));
            computeStatistics();
        }
        else
            System.out.println(">> tried to send non Number and non Boolean value to " + toString());
    }


// number ----------------------------------------------------------------------------------------------------

  /** variable to store the number object */
  Number number;

  /** Sets the number object received by a propertyChange */
  public void setNumber(Number newNumber) {
      number = newNumber;
  }

  /** Gets the number object */
  public Number getNumber() { return number; }



// -----------------------------------------------------------------------------------------------------------
// statistic values ------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------


// min  ------------------------------------------------------------------------------------------------------

  /** variable to hold the minimum value */
  protected double min;

  /** Gets the minimum value */
  public double getMin () {
     return min;
  }
  /** sets the minimum value */
  protected void setMin (double newMin) {
     min = newMin;
        // FOR LATER USE: ---> add signaling property change here
  }


// max  ------------------------------------------------------------------------------------------------------

  /** variable to hold the maximum value */
  protected double max;

  /** Gets the maximum value */
  public double getMax () {
        return max;
  }

  /** Sets the maximum value */
  protected void setMax (double newMax) {
        max = newMax;
        // FOR LATER USE: ---> add signaling property change here
  }

// startTime  ------------------------------------------------------------------------------------------------

  /** variable to hold the start time of the statistics computation */
  protected double startTime;

  /** Gets the start time of the statistics computation  */
  public double getStartTime () {
      return startTime;
  }

  /** Sets the start time of the statistics computation  */
  protected void setStartTime (double newTime) {
     startTime = newTime;
  }


// time of last event ----------------------------------------------------------------------------------------

  /** variable to hold the time of last value change  */
  protected double tL;

  /** Gets the tL value */
  protected double getTL () {
     return tL;
  }

  /** Sets the tL value */
  protected void setTL (double newTL) {
     tL = newTL;
  }


// total elapsed time ----------------------------------------------------------------------------------------

  /** variable to hold the total elapsed time for the statistics  */
  protected double totalElapsedT = 0;

  /** Gets the total elapsed time for the statistics */
  protected double getTotalElapsedT () {
     return totalElapsedT;
  }

  /** Sets the total elapsed time for the statistics */
  protected void setTotalElapsedT (double newTotalElapsedT) {
     totalElapsedT = newTotalElapsedT;
  }


// compute statistics ----------------------------------------------------------------------------------------

  /** Computes the statistics based on the received number values */
  protected void computeStatistics() {
    double val = number.doubleValue();
    if (getMax() < number.doubleValue())
        setMax(number.doubleValue());
    if (getMin() > number.doubleValue())
        setMin(number.doubleValue());
    double elapsedT = getTime() - getTL();
    setTotalElapsedT (getTotalElapsedT() + elapsedT); 
  }



// reset -----------------------------------------------------------------------------------------------------

  /** reset the output bean */
  public void reset () {
    super.reset(); 
    clearStatistics();
  }

  /** Clears the statistics */
  protected void clearStatistics() {
    setStartTime(getTime());
    setMax(-Double.MAX_VALUE);
    setMin(Double.MAX_VALUE);
    setTL (getTime());
    setTotalElapsedT (0);
  }

}

