package mosaic.sim.kernel.variable;

import java.awt.Graphics;
import java.awt.Color;
import mosaic.sim.kernel.AbstractModelComponent;

/**
*
*/

public class BasicConnection
   extends AbstractModelComponent
   implements Connection{

   protected Variable target;
   protected Variable domain;

   protected boolean continuousMode=false;

   /** Tells whether the connection is established or not.*/
   protected boolean connected=false;

   public BasicConnection() {
      super("connection");
      } // BasicConnection;

   public BasicConnection(String name) {
      super(name);
      } // BasicConnection


   public BasicConnection(Variable domain, Variable target){
     super("connection");
     setDomain(domain);
     setTarget(target);
   }

    public boolean isContinuousMode(){
        return continuousMode;
    }

    protected void getContinuous(){
        if (!continuousMode){
            continuousMode=true;
            fireModelComponentGotDiscrete();
            if (enabled && target!=null && target instanceof DoubleVariable){
                ((DoubleVariable)target).inCouplerGotContinuous();
            }
        }
    }

    protected void getDiscrete(){
        if (continuousMode){
            continuousMode=false;
            fireModelComponentGotDiscrete();
            if (enabled && target!=null && target instanceof DoubleVariable){
                ((DoubleVariable)target).inCouplerGotDiscrete();
            }
        }
    }

    public void domainGotDiscrete(){
        getDiscrete();
    }

    public void domainGotContinuous(){
        getContinuous();
    }

   /** connects two Variables */
   protected final void connect() throws VariableException{
     if (!enabled) return; // do nothing
     // target already defined
//     if (!connected && target.getInCoupler()!=null) throw(new VariableException("Variable already defined ", target));
     // test cycle of connections
     if (domain.getDefinedBy()==target) throw(new VariableException("Variable connection produces cycle ", target));
//     Variable newDefining;
     // set new source
     target.setDefinedBy(domain.getDefinedBy());
     // The connection has been established.
     connected=true;
   }


   /** disconnects two Variables */
   protected final void disconnect(){
     if (!connected) return; // There is nothing to do.
     target.setDefinedBy(target); // remove source from target
     try{
       target.setInCoupler(null);
     }
     catch(VariableException e){}
     domain.removeOutCoupler(this);
     connected=false; // The connection has been removed
   }

   /**
   * Set the domain variable
   */
   public final void setDomain (Variable domain){
     if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
       getSimApplication().getIntegrator().stateEventOccurred = true;
       return;
     }
     if (this.domain==domain) return; // nothing changed
     if (enabled && this.domain!=null) this.domain.removeOutCoupler(this); //remove this from old source
     this.domain=domain; // set new domain
     if (enabled && domain!=null){
       domain.addOutCoupler(this); // add this to new source
       if (target!=null)
         try{connect(); // create new connection
         }catch(VariableException e){if (getSimApplication()!=null) printErrorOutput(e.getMessage());}
     }
     if (domain instanceof DoubleVariable){
        if (domain==null) getDiscrete();
        if (((DoubleVariable)domain).isContinuous()) getContinuous();
     }else{
        getDiscrete();
     }
     fireModelComponentChanged(); // inform listeners
   }

   /**
   * Set the target variable
   */
   public final void setTarget (Variable target){
     if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
       getSimApplication().getIntegrator().stateEventOccurred = true;
       return;
     }
     if (this.target==target) return; // nothing changed
     try{
       if (enabled && this.target!=null) this.target.setInCoupler(null);
     }catch(VariableException e){if (getSimApplication()!=null) printErrorOutput(e.getMessage());}
     this.target=target; // set new target
     if (enabled && target!=null){
       try{
         target.setInCoupler(this);  // set new domain
         if (domain!=null) connect(); // create new connection
       }catch(VariableException e){if (getSimApplication()!=null) printErrorOutput(e.getMessage());}
     }
     fireModelComponentChanged(); // inform listeners
   }


   /**
   * Get the domain variable
   */
   public final Variable getDomain () {
      return (domain);
   }

   /**
   * Get the domain variable
   */
   public final Variable getTarget () {
      return (target);
   }

   public void setEnabled(boolean enabled){
     if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
       getSimApplication().getIntegrator().stateEventOccurred = true;
       return;
     }

     super.setEnabled(enabled); // supercall to BasicModel
     if (enabled){
       if (connected) return;
       try{
         target.setInCoupler(this);
         domain.addOutCoupler(this);
         connect();
       } // try to connect
       catch(VariableException e){
            printErrorOutput(e.toString());
       }
     }
     else disconnect(); // remove connection
   }

   public final void definedByChange(Variable d){
//     if (!enabled) return; // nothing to do
     if (target!=null){
       try{
         connect();
       }
       catch(VariableException ex) {
            printErrorOutput(ex.toString());
       }
     }
   }

   public final void propagateValueChange(Variable d){
     if (!connected) return;
     target.signalVariableChange();
   }

   public final void domainTypeChange(Variable d){
//     if (!enabled) return;
        fireModelComponentChanged(); // inform listeners
   }

   public final void targetTypeChange(Variable d){
//     if (!enabled) return;
        fireModelComponentChanged(); // inform listeners
   }

   protected void finalize(){
     try{
       super.finalize();
     }catch(Throwable t){}
     disable();
   }

    /**
    * Paints the connection.
    * Paints an arrow and labels it with name of domain and target.
    * @param g      the graphics context of the component
    */
    public void paintComponent(Graphics g) {
      int h = getSize().height;
      int w = getSize().width;
      if (enabled) g.setColor(Color.black);
      else g.setColor(Color.gray);
      g.drawLine(3, h - 6, w-3, h - 6);
      g.drawLine(w-8, h - 6 + 4, w-3, h - 6);
      g.drawLine(w-8, h - 6 - 4, w-3, h - 6);
      String lbl = "";
      if (domain != null) {
        lbl = domain.getLabel();
      }
      lbl = lbl + " -> ";
      if (target != null) {
        lbl = lbl + target.getLabel();
      }
      g.drawString(lbl,2,h-7);
    }

}
// end of BasicConnection.java
