package mosaic.sim.kernel.variable;

import mosaic.sim.kernel.control.*;
import java.util.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.simevent.*;

/**
*  Derived from Variable (support VariableChangeListeners)
*  Holds a double value
*/

public class DoubleVariable
   extends mosaic.sim.kernel.variable.AbstractVariable
   {
   /** Numeric value represented by this <code> DoubleVariable </code>. */
   protected double value;

   /** counter used to assert a minimum update rate */
   protected int counter = 0;

   /** threshold, to assert update if value changes significantly */
   protected double threshold = 0.1;

   /** holds the current mode of the variable. */
   protected boolean continuous=false;


  /** Default constructor for BooleanVariable*/
   public DoubleVariable() {
        this("doubleVariable", VARIABLE, "", "", "", 0.0);
   }

  /** Constructor for DoubleVariable initializing model owner and value. */
   public DoubleVariable (double value) {
        this("doubleVariable", VARIABLE, "", "", "", value);
   }

  /** Constructor for DoubleVariable initializing model owner, name, and value. */
   public DoubleVariable (String name, double value) {
        this(name, VARIABLE, "", "", "", value);
   }

  /** Constructor for DoubleVariable initializing model owner, name, variable type, and value. */
   public DoubleVariable (String name, int type, double value) {
        this(name, type, "", "", "", value);
   }

  /** Constructor for DoubleVariable initializing model owner, name,
  * description, unit of measurement, quantity, displayUnit, variable type, and value. */
  public DoubleVariable(String name, int type,
                    String unit, String quantity, String displayUnit, double value ) {
      super(name, type, unit, quantity, displayUnit);
      initializeMembers(value);
   }


   /**
   * Initializes the value and the previous value.
   */
   private void initializeMembers(double value) {
      this.value = value;
      this.previousValue = Double.NaN;
   } // initializeMembers


    public boolean isContinuous(){
        return continuous;
    }

    public void inCouplerGotContinuous(){
        getContinuous();
    }

    protected final void getContinuous(){
        if (!continuous){
            printDebugOutput(label+" continuous");
            continuous=true;
            fireModelComponentGotContinuous();
            if (outCouplers.size()>0){
                Iterator i=outCouplers.iterator();
                Coupler outCoupler;
                while (i.hasNext()){
                    outCoupler=(Coupler)i.next();
                    outCoupler.domainGotContinuous();
                }
            }
        }
    }

    public void inCouplerGotDiscrete(){
        if (!isTypeState()){
            getDiscrete();
        }
    }

    protected final void getDiscrete(){
        if (continuous){
            printDebugOutput(label+" discrete");
            continuous=false;
            fireModelComponentGotDiscrete();
            if (outCouplers.size()>0){
                Iterator i=outCouplers.iterator();
                Coupler outCoupler;
                while (i.hasNext()){
                    outCoupler=(Coupler)i.next();
                    outCoupler.domainGotDiscrete();
                }
            }
        }
    }

    public final void setInCoupler(mosaic.sim.kernel.variable.Coupler in) throws VariableException{
        super.setInCoupler(in);
        if (!isTypeState()){
            if (!continuous && in!=null && in.isContinuousMode()){
                getContinuous();
            }else if(continuous && (in==null || !in.isContinuousMode())){
                getDiscrete();
            }
        }
    }

   /**
   *
   */
   public final double getValue() {
     return ((DoubleVariable)definedBy).value;
   } // getValue

   /**
   *
   */
   public final Object getValueObject(){
     return new Double(getValue());
   }

   /**
   *
   */
   public void setValueObject(Object o)
      { try {setValue( ((Double)o).doubleValue()); }
        catch( java.lang.ClassCastException e) {
            printErrorOutput("Tried to send non Double Object as Value to DoubleVariable");
      }
   }


    /**
    * Sets the value of this double variable
    */
    public final void setValue(double value) {
        if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
            getSimApplication().getIntegrator().stateEventOccurred = true;
            return;
        }
        if (value >= minValue && value <= maxValue) {
            this.value = value;
            if (! continuous || ! getSimApplication().getIntegrator().isInIntegrationStep) {
                signalVariableChange();
            }
        }
    }

   /** */
   protected DoubleVariable derivative;

   /**
   *
   */
   public final DoubleVariable getDerivative()
      { return derivative; }

   /**
   *
   */
    public final void setDerivative(DoubleVariable derivative){
        if (this.derivative==derivative) return;
        this.derivative = derivative;
        if (derivative==null){
            setTypeVariable();
            if (inCoupler==null||!inCoupler.isContinuousMode()) getDiscrete();
        }else{
            setTypeState();
            getContinuous();
        }
    }


// previous value and time ------------------------------------------------------------------------------

   /** time of the last continuous change */
   protected double previousTime;

   /** value at the last continuous change */
   protected double previousValue;


   /**
   *  Gets the value at the last continuous state change.
   */
   public final double getPreviousValue() {
     return ((DoubleVariable)definedBy).previousValue;
      } // getPreviousValue

   /**
   *  Gets the time at the last continuous state change.
   */
   public final double getPreviousTime() {
     return ((DoubleVariable)definedBy).previousTime;
      } // getPreviousValue

   /**
   *  Gets the value the before the next continuous state change.
   */
   public void saveToPrevious() {
      if (definedBy ==this) {
         previousTime = changedTime;
         previousValue = value;
      }
   }

// min value ---------------------------------------------------------------------------------------

    /** minimum value allowed in this variable */
    double minValue = Double.NEGATIVE_INFINITY;

    /** Gets the minimum value allowed in this variable */
    public double getMinValue() {
        return minValue;
    }

    /** Sets the minimum value allowed in this variable */
    public void setMinValue(double minValue) {
        this.minValue = minValue;
    }

// mas value ---------------------------------------------------------------------------------------

    /** minimum value allowed in this variable */
    double maxValue = Double.POSITIVE_INFINITY;

    /** Gets the minimum value allowed in this variable */
    public double getMaxValue() {
        return maxValue;
    }

    /** Sets the minimum value allowed in this variable */
    public void setMaxValue(double maxValue) {
        this.maxValue = maxValue;
    }



// -------------------------------------------------------------------------------------------------
// overwrite method processSimEvent from interface InputEventListener  -----------------------------
// -------------------------------------------------------------------------------------------------

  /** Reacts to SimEvents from SimEventComponents.
   *  The events should be triggered by a inputEventComponent and
   *  return Number-Objects.
   */
  public void processSimEvent (SimEvent evt){
  Object obj;
    if ((evt.getSource()) instanceof InputEventComponent){
      obj = ((InputEventComponent)(evt.getSource())).getInput();
      if (obj instanceof Number)
        this.setValue( ((Double)(obj)).doubleValue() );
    }
  }

} // class DoubleVariable

// end of DoubleVariable.java



