/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: PLSynapse.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.7 $
 *  Last changed at: $Date: 2002/04/24 19:41:31 $
 *  Last changed by: $Author: Rene $
 *****************************************************************************/

package mosaic.sim.neuron;

/** This class represents synapses with piecewise linear response functions.
 */
public abstract class PLSynapse extends Synapse {
	/** The weight of the synapse, i.e. its strenght. It defines how much the
	 *  post-synaptic neuron potential changes when a spike is received by this
	 *  synapse.
	 *  This parameter will typically be changed during the learning phase.
	 */
	protected float weight;

	/** The delay of the synapse. It defines the delay between the receipt of
	 *  the incoming spike on this synapse and the time when the post-synaptic
	 *  neuron potential starts to change due to this spike.
	 *  This parameter will typically be changed during the learning phase.
	 */
	protected float delay;

	/** Initializes the sender and receiver objects (by calling the super()
	 *  constructor) and the weight and delay values with default.
	 */
	public PLSynapse(NeuronOutput pre, Neuron post) {
		super(pre, post);
		Debug.assertCond(pre != null, "PLSynapse.PLSynapse",
			"pre can not be null");
		Debug.assertCond(post != null, "PLSynapse.PLSynapse",
			"post can not be null");
		weight = (float) 1.0;
		delay = (float) 0.0;
	}

	/** Calls the constructor with two parameters, but also allows to specify
	 *  the synapse weight.
	 */
	public PLSynapse(NeuronOutput pre, Neuron post, float weight) {
		this(pre, post);
		setWeight(weight);
	}

	/** Calls the constructor with three parameters, but also allows to specify
	 *  the synapse delay.
	 */
	public PLSynapse(NeuronOutput pre, Neuron post,
						float weight, float delay) {
		this(pre, post, weight);
		setDelay(delay);
	}

	/** Calls the constructor with four parameters, but also allows to specify
	 *  the learning method.
	 */
	public PLSynapse(NeuronOutput pre, Neuron post,
						float weight, float delay,
						LearningMethod learningMethod) {
		this(pre, post, weight, delay);
		setLearningMethod(learningMethod);
	}

	/** This implementation of Synapse.getResponseParameters returns an object
	 *  of type PLFunction, i.e. 2 one-dimensional arrays describing the
	 *  piecewise linear response function of this synapse. The weight and
	 *  delay parameters are applied to the function before returning it here,
	 *  so they do not need to be taken into account somewhere else.
	 *  @see PLFunction
	 */
	public Object getResponseParameters() {
		/* special case: if the weight is zero, just return nothing - this is
		   an optimization */
		if (weight == 0)
		   return new PLFunction(0);
		/* first of all, get the gradient parameters and store them locally
		   so that the get* functions don't have to be called too often */
		PLFunction standardResponse = getStandardResponse();
		Debug.assertCond(standardResponse.gradients.length ==
			standardResponse.gradientTimes.length,
			"PLSynapse.getResponseParameters",
			"gradient and time arrays must have equal length");
		PLFunction response = new PLFunction(standardResponse.gradients.length);
		// the times are shifted by the synpapse delay
		for (int i=0; i<response.gradients.length; i++)
			response.gradients[i] = standardResponse.gradients[i] * weight;
		for (int i=0; i<response.gradients.length; i++)
			response.gradientTimes[i] = standardResponse.gradientTimes[i] +
				delay;
		return response;
	}

	/** Gets the currently set synaptic weight. */
	public float getWeight() {
		return weight;
	}

	/** Sets the synaptic weight. */
	public void setWeight(float weight) {
		Debug.assertCond(! Float.isInfinite(weight) && ! Float.isNaN(weight),
			"PLSynapse.PLSynapse", "weight can not be Infinity or NaN");
		this.weight = weight;
	}

	/** Gets the currently set synaptic delay. */
	public float getDelay() {
		return delay;
	}

	/** Sets the synaptic delay. */
	public void setDelay(float delay) {
		Debug.assertCond(delay >= 0.0 && ! Float.isInfinite(delay),
			"PLSynapse.PLSynapse", "delay must be >= 0");
		this.delay = delay;
	}

	/** getResponseParameters uses this method to retrieve the function
	 *  gradients used for constructing the response object. The gradients
	 *  should have the standard values for this synapse without taking the
	 *  weight and delay parameters into account.
	 */
	protected abstract PLFunction getStandardResponse();
}
