/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: ImageInput.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.4 $
 *  Last changed at: $Date: 2002/05/19 17:20:49 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron.netinputs;

import java.io.*;
import java.awt.image.*;
import javax.media.jai.*;
import javax.media.jai.iterator.*;
import mosaic.sim.neuron.*;

/** This class reads an image and converts it to either a one-dimensional vector
 *  of dimension height*width or a two-dimensional matrix of dimension
 *  [height, width].
 *  For inputting a one-dimensional vector to the network, this class uses
 *  the interface VectorInput; for inputting a two-dimensional matrix, it uses
 *  MatrixInput. Objects implementing these interface have to
 *  be given in the constructor when initializing.
 *
 *  This class needs the JAI (Java Advanced Imaging) library to do its work.
 *  @see VectorInputCoder
 *  @see MatrixInputCoder
 */
public class ImageInput {
	/** The dimension of the network input. */
	protected int dim = 0;
	/** This object is used as the coder when using one-dimensional (vector)
	 *  input.
	 */
	protected VectorInputCoder coderOneDim = null;
	/** This object is used as the coder when using two-dimensional (matrix)
	 *  input.
	 */
	protected MatrixInputCoder coderTwoDim = null;

	/** The image. */
	protected PlanarImage image = null;

	/** This constructor initializes the class for one-dimensional or
	 *  two-dimensional input.
	 *  Note that the input can not be done until the coder has been set to
	 *  an object that implements the suitable coder class.
	 *  The network input can be started by calling startInput
	 *
	 *  @see #setCoder
	 *  @see #startInput
	 *
	 *  @param filename The filename of the image file that should be used.
	 *  @param dim The dimension of the network input (1 or 2)
	 */
	public ImageInput(String filename, int dim) throws CodingException,
		   FileNotFoundException {
		if (dim != 1 && dim != 2)
		   throw new CodingException("Only one or two dimensional input is supported.");
		this.dim = dim;

		// now get the image from the file
		File f = new File(filename);
		if (! f.exists() || ! f.canRead())
		   throw new FileNotFoundException();
		image = JAI.create("fileload", filename);
	}

	/** Returns the height of the loaded image. */
	public int getHeight() {
		return image.getHeight();
	}

	/** Returns the width of the loaded image. */
	public int getWidth() {
		return image.getWidth();
	}

	/** Returns the color model of the loaded image. */
	public ColorModel getColorModel() {
		return image.getColorModel();
	}

	/** Returns the sample model of the loaded image. */
	public SampleModel getSampleModel() {
		return image.getSampleModel();
	}

	/** Initializes the Spike coder. */
	public void setCoder(VectorInputCoder coder) throws CodingException {
		if (dim != 1)
		   throw new CodingException("A VectorInput coder can only be used for one dimensional input.");
		coderOneDim = coder;
	}

	public void setCoder(MatrixInputCoder coder) throws CodingException {
		if (dim != 2)
		   throw new CodingException("A MatrixInput coder can only be used for two dimensional input.");
		coderTwoDim = coder;
	}

	/** This method really starts the network input and should only be called
	 *  after the coder has been initialized with setCoder().
	 */
	public void startInput() throws CodingException {
		if ((dim == 1 && coderOneDim == null) ||
			(dim == 2 && coderTwoDim == null))
		   throw new CodingException("Network input can not be started - set a coder first.");

		// used to access the source image
		RandomIter iter = RandomIterFactory.create(image, null);

		double[] vector = null;
		double[][] matrix = null;

		if (dim == 1)
		   vector = new double[getHeight() * getWidth()];
		else
			matrix = new double[getHeight()][getWidth()];

		// here, as a simplification, we only use band 0 of each loaded image
		for (int i=0; i<getHeight(); i++) {
			for (int j=0; j<getWidth(); j++) {
				if (dim == 1)
				   vector[i*getWidth() + j] = iter.getSample(j, i, 0);
				else
				   matrix[i][j] = iter.getSample(j, i, 0);
			}
		}

		// and now - the network input
		if (dim == 1)
		   coderOneDim.enterVector(vector);
		else
			coderTwoDim.enterMatrix(matrix);
	}
}
