/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: PLNeuronInternalView.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.19 $
 *  Last changed at: $Date: 2002/05/16 20:43:35 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron.visualisation;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.variable.*;
import mosaic.sim.neuron.*;
import java.beans.PropertyChangeListener;
import java.io.Serializable;
import java.beans.PropertyChangeEvent;

/** This class implements a view for the potential and threshold functions of
 *  a neuron, thus its internal states.
 */
public class PLNeuronInternalView extends ScrollingView
		implements VariableChangeListener {

	/** If set to true, the view displays the neuron potential. */
	protected boolean displayPotential;
	/** If set to true, the view displays the neuron threshold. */
	protected boolean displayThreshold;
	/** A reference to the neuron potential object. */
	protected PLPotential potential = null;
	/** A reference to the neuron threshold object. */
	protected PLThreshold threshold = null;

	/** The currently used potential function. */
	protected LinkedList potentialFunction = null;
	/** The currently used threshold function. */
	protected LinkedList thresholdFunction = null;

	/** This method draws a piecewise linear function which has to be given as
	 *  a linear list containing points in the world (not graphics) space as
	 *  Point2D.Double objects.
	 *
	 *  This method is synchronized so that scrollToSimulationTime can not be
	 *  called in between, making previous calls to worldToGraphics invalid
	 *  (because of the changed time frame, the previously computed values
	 *  would have to be computed again).
	 *
	 *  @param plFunction The piecewise linear function to draw.
	 *  @param g The graphics context to draw into.
	 *  @param pastColor The color to be used for segments that are in the past.
	 *  @param futureColor The color to be used for segments that are in the
	 *                     future.
	 */
	synchronized protected void drawPLFunction(LinkedList plFunction,
									Graphics g,
									Color pastColor, Color futureColor) {
		ListIterator iterator = plFunction.listIterator();
		Point lastGraph = null;
		Point2D.Double lastWorld = null;
		boolean currentTimePassed = false;
		double currentTime = SimApplication.getCurrentSimApplication().getTime();

		g.setColor(pastColor);
		while (iterator.hasNext()) {
			Point2D.Double nextWorld = (Point2D.Double) iterator.next();
			// only paint lines that will be at least partially visible
			if (nextWorld.x >= tmin &&
					(lastWorld == null || lastWorld.x <= tmax)) {
				Point nextGraph = worldToGraphics(nextWorld.x, nextWorld.y);
				Debug.assertCond(lastGraph == null || nextGraph.x >= lastGraph.x,
					"PLNeuronInternalView.drawPLFunction",
					"Time steps have to be >= 0: last x = " +
					((lastGraph != null) ? String.valueOf(lastGraph.x) :
					"null") +
					", next x = " +
					nextGraph.x +
					", last time = " +
					((lastWorld != null) ? String.valueOf(lastWorld.x) :
					"null" ) +
					", next time = " +
					nextWorld.x);
				if (lastGraph != null) {
					if (nextWorld.x >= currentTime && !currentTimePassed) {
						/* If the gradient is less than infinite, then the line
						   which is to be drawn will cross the current time
						   point. Therefore it is necessary to calculate a
						   point in between (at the current time) to be able to
						   switch colors at this point. */
						if (nextGraph.x > lastGraph.x) {
							Point temp = worldToGraphics(currentTime, 0);
							temp.y = (int) (lastGraph.y +
								(nextGraph.y - lastGraph.y) *
								(temp.x - lastGraph.x) /
								(nextGraph.x - lastGraph.x));
							g.drawLine(lastGraph.x, lastGraph.y, temp.x, temp.y);
							lastGraph = temp;
						}
						g.setColor(futureColor);
						currentTimePassed = true;
					}
					g.drawLine(lastGraph.x, lastGraph.y,
						nextGraph.x, nextGraph.y);
				}
				lastGraph = nextGraph;
				lastWorld = nextWorld;
			}
		}
	}

	/** Creates a PLNeuronInternalView and initializes it.
	 *  @param title The title that should be set for the view.
	 *  @param displayPotential The view only displays a neuron potential if
	 *                          this is set to true
	 *  @param displayThreshold The view only displays a neuron threshold if
	 *                          this is set to true
	 *  @see #displayPotential
	 *  @see #displayThreshold
	 */
	public PLNeuronInternalView(String title,
								boolean displayPotential,
								boolean displayThreshold) {
		this.displayPotential = displayPotential;
		this.displayThreshold = displayThreshold;

		if (displayPotential)
			potentialFunction = new LinkedList();
		if (displayThreshold)
			thresholdFunction = new LinkedList();

		this.setBackground(Color.black);
		/*this.setBorder(BorderFactory.createTitledBorder(
			   BorderFactory.createLineBorder(Color.green), title));*/
	}

	/** Overrides the default and draws the piecewise linear functions for the
	 *  neuron threshold and potential.
	 */
	synchronized public void paintComponent(Graphics g) {
		super.paintComponent(g);

		if (displayPotential && potential != null) {
			drawPLFunction(potentialFunction, g, Color.green, Color.lightGray);
		}
		if (displayThreshold && threshold != null) {
			drawPLFunction(thresholdFunction, g, Color.red, Color.lightGray);
		}
		if (displayThreshold && displayPotential &&
				potential != null && threshold != null) {
			// draw the next scheduled event for this object
			double nextFireTime = threshold.calculateNextFireTime(potential);
			if (nextFireTime >
					SimApplication.getCurrentSimApplication().getTime()) {
				Point from = worldToGraphics(nextFireTime, tmin),
					  to = worldToGraphics(nextFireTime, tmax);
				g.setColor(Color.gray);
				g.drawLine(from.x, from.y, to.x, to.y);
			}
		}
	}

	/** Implementation of VariableChangeListener:
	  * This method is called by the simulation framework whenever a
	  * variable that this object is listening on changes. It is used for
	  * receiving the change events when the potential or threshold objects
	  * change.
	  */
	synchronized public void variableChange (VariableChangeEvent e) {
		//System.out.print("PLNeuronInternalView.variableChange called ");

		Object value = e.getValue();

		//System.out.println("with object " + value.getClass().toString());

		/* since this view can display both a potantial and a threshold graph,
		   we need to handle both types of change events here. */
		if (value instanceof PLPotential) {
			if (displayPotential) {
				potential = (PLPotential) value;
				// optimization (and we depend on a length of at least one below...)
				if (potential.getChanges().gradients.length == 0)
				   return;

			///
				/*ListIterator old = potentialFunction.listIterator();
				System.out.println("*****************");
				System.out.print("Received potential variable change event. Old was: ");
				while (old.hasNext()) {
					Point2D.Double p;
					p = (Point2D.Double) old.next();
					System.out.print(p.x + "/" + p.y + ", ");
				}
				System.out.println();
				System.out.println("Received potential is: ");
				for (int i=0; i<potential.getNumSegments(); i++)
					System.out.println(potential.getChanges().gradients[i] +
						" at " + potential.getChanges().gradientTimes[i]);*/
			///

				ListIterator old = potentialFunction.listIterator();
				// double val = potential.getChanges().getEndTime(potential.getLastValue();
				double val = 0;
				double currentTime = SimApplication.getCurrentSimApplication().getTime();
				int numSeg = potential.getNumSegments();
				Point2D.Double p, last = new Point2D.Double();
				/* skip all the segments that can not change anymore (which are
				   before the current simulation time and not contained in the
				   current neuron potential) */
				while (old.hasNext()) {
					p = (Point2D.Double) old.next();
					if (p.x >= potential.getChanges().gradientTimes[0]) {
						if (old.hasPrevious())
							old.previous();
						else
							// this is for the first element
							old = potentialFunction.listIterator();
						break;
					}
				}
				/* remove all the segments that are contained in
				   the current neuron potential and take them from there */
				while (old.hasNext()) {
					old.next();
					old.remove();
				}
				// now the part that has been skipped has to end with a value of 0
				Debug.assertCond(potentialFunction.size() ==  0 ||
					((Point2D.Double) potentialFunction.getLast()).y == 0,
					"PLNeuronInternalView.variableChange",
					"Past potential does not terminate with a value of 0.");

				for (int i=0; i < numSeg; i++) {
					p = new Point2D.Double();
					p.x = potential.getChanges().gradientTimes[i];
					p.y = val;
					potentialFunction.add(p);
					if (i < numSeg - 1)
						val += potential.getChanges().gradients[i] *
							(potential.getChanges().gradientTimes[i+1] -
							 potential.getChanges().gradientTimes[i]);
				}
				//last.x = potential.getChanges().getEndTime(potential.getLastValue(), potential.getNumSegments());
				last.x = potential.getChanges().getEndTime(0, potential.getNumSegments());
				last.y = 0;
				potentialFunction.add(last);

				///
				/*old = potentialFunction.listIterator();
				System.out.print("New is: ");
				while (old.hasNext()) {
					p = (Point2D.Double) old.next();
					System.out.print(p.x + "/" + p.y + ", ");
				}
				System.out.println();*/
				///
			}
		}
		else if(value instanceof PLThreshold) {
			if (displayThreshold) {
				threshold = (PLThreshold) value;

			///
				/*ListIterator old = thresholdFunction.listIterator();
				System.out.println("*****************");
				System.out.print("Received threshold variable change event. Old was: ");
				while (old.hasNext()) {
					Point2D.Double p;
					p = (Point2D.Double) old.next();
					System.out.print(p.x + "/" + p.y + ", ");
				}
				System.out.println();*/
			///


				ListIterator old = thresholdFunction.listIterator();
				double currentTime = SimApplication.getCurrentSimApplication().getTime();
				Point2D.Double p, q;
				/* skip all the segments that can not change anymore (which are
				   fully before the current simulation time) */
				/*while (old.hasNext()) {
					p = (Point2D.Double) old.next();
					if (p.x >= currentTime) {
						if (old.hasPrevious())
							old.previous();
						else
							// this is for the first element
							old = potentialFunction.listIterator();
						break;
					}
				}
				// remove all the future segments
				while (old.hasNext()) {
					old.next();
					old.remove();
				}*/
				// remove all points that are in the future
				if (! thresholdFunction.isEmpty()) {
					p = (Point2D.Double) thresholdFunction.getFirst();
					while (((Point2D.Double) thresholdFunction.getLast()).x >
							 currentTime) {
						p = (Point2D.Double) thresholdFunction.removeLast();
					}
					/* insert a point for painting the last segment until the
					   current simulation time */
					q = (Point2D.Double) thresholdFunction.getLast();
					if (p != q)
					   thresholdFunction.add(new Point2D.Double(currentTime,
						   q.y + (p.y - q.y) / (p.x - q.x) * (currentTime - q.x)));
				}
				else {
					thresholdFunction.add(new Point2D.Double(0,
						threshold.getQuiescentValue()));
					thresholdFunction.add(new Point2D.Double(currentTime,
						threshold.getQuiescentValue()));
				}

				switch (threshold.getState()) {
					case PLThreshold.STATE_QUIESCENT:
						thresholdFunction.add(new Point2D.Double(currentTime,
							threshold.getQuiescentValue()));
						thresholdFunction.add(new Point2D.Double(
							Double.POSITIVE_INFINITY,
							threshold.getQuiescentValue()));
						break;
					case PLThreshold.STATE_ABSOLUTE_REFRACTION:
						thresholdFunction.add(new Point2D.Double(currentTime,
											  Double.POSITIVE_INFINITY));
						break;
					case PLThreshold.STATE_RELATIVE_REFRACTION:
						double val = threshold.getRelativeRefractionStartValue();
						double starttime = threshold.getLastFireTime() +
								threshold.getAbsoluteRefractionTime();
						int i;
						PLFunction changes = threshold.getRelativeRefraction();
						for (i=0; i<changes.gradients.length-1; i++) {
							thresholdFunction.add(new Point2D.Double(
								starttime + changes.gradientTimes[i],
								val));
							val += changes.gradients[i] *
								(changes.gradientTimes[i+1] -
								changes.gradientTimes[i]);
						 }
						thresholdFunction.add(new Point2D.Double(
							starttime + changes.gradientTimes[i],
							val));
						break;
					default:
						Debug.assertCond(false, "PLNeuronInternalView.variableChange",
							"Received threshold object with unknown state.");
				}

				///
				/*old = thresholdFunction.listIterator();
				System.out.print("New is: ");
				while (old.hasNext()) {
					p = (Point2D.Double) old.next();
					System.out.print(p.x + "/" + p.y + ", ");
				}
				System.out.println();*/
				///
			}
		}
		else
			Debug.assertCond(false, "PLNeuronInternalView.variableChange",
				"Received unknown variable change event.");

		updateView();
	}

	/** Implementation of VariableChangeListener:
	 *  This implementation does nothing.
	 */
	public void propertyChange (PropertyChangeEvent e) {
	}

	/** Implementation of VariableChangeListener:
	 *  This implementation does nothing.
	 */
	public void addListenerVariable (mosaic.sim.kernel.variable.Variable v) {
	}

	/** Implementation of VariableChangeListener:
	 *  This implementation does nothing.
	 */
	public void removeListenerVariable (mosaic.sim.kernel.variable.Variable v) {
	}

	/** Implementation of VariableChangeListener:
	 *  This implementation always returns 0.
	 */
	public int getPriority() {
		return 0;
	}
}
