/*
 * %LogOutput%
 *
 * copyright 2000, JKU Linz, Austria
 */

package mosaic.util;

import java.io.*;
import java.util.*;


/**
 * Class LogOutput is a general class to allow controlled output of log, error and debug messages.
 * It maintains a map of message types (e.g. "error", "debug", "simulation" etc.) to PrintStreams
 * where the messages of this type should appear together with outputLevels which allow to
 * use different levels of output.
 */
public class LogOutput {

	/** Class to hold a pair of print stream and output level */
	private class StreamLevelPair {

		StreamLevelPair(PrintStream printStream, int outputLevel) {
			this.printStream = printStream;
			this.outputLevel = outputLevel;
		}

		StreamLevelPair() {
			this.printStream = System.out;
			this.outputLevel = 1;
		}

		PrintStream printStream;
		int outputLevel;
	}


	/** Default constructor with no arguments */
	public LogOutput () {
		setLogStream("log", System.out, defaultLog.outputLevel);
		setLogStream("error", System.err, defaultLog.outputLevel);
		setLogStream("debug", System.out, defaultLog.outputLevel);
		setLogStream("simulation", System.out, defaultLog.outputLevel);
	}

	/** Default stream and output level */
	private StreamLevelPair defaultLog = new StreamLevelPair(System.out, 1);

	/** Map of message types to output streams. */
	private HashMap logStreams = new HashMap();

	/** Map of message types to output streams. */
	private HashMap outputLevels = new HashMap();

	/** Sets ouput stream for message type. */
	public void setLogStream(String messageType, PrintStream logStream) {
		logStreams.put(messageType, new StreamLevelPair(logStream, defaultLog.outputLevel));
	}

	/** Sets ouput stream and output level for message type. */
	public void setLogStream(String messageType, PrintStream logStream, int outputLevel) {
		logStreams.put(messageType, new StreamLevelPair(logStream, outputLevel));
	}

	/** Sets ouput level for message type. */
	public void setOutputLevel(String messageType, int outputLevel) {
		logStreams.put(messageType, new StreamLevelPair(getLogStream(messageType), outputLevel));
	}

	/** Sets default ouput stream and output level for message type. */
	public void setDefaultLogStream(PrintStream logStream) {
		defaultLog.printStream = defaultLog.printStream;
	}

	/** Sets default ouput stream and output level for message type. */
	public void setDefaultLogStream(PrintStream logStream, int outputLevel) {
		defaultLog.printStream = defaultLog.printStream;
		defaultLog.outputLevel = outputLevel;
	}

	/** Sets ouput level for message type. */
	public void setDefaultOutputLevel(int outputLevel) {
		defaultLog.outputLevel = outputLevel;
	}

	/** Gets the output stream for the message type. */
	public PrintStream getLogStream(String messageType) {
		StreamLevelPair pair = ((StreamLevelPair)logStreams.get(messageType));
		if (pair != null) {
			return pair.printStream;
		} else {
			return defaultLog.printStream;
		}
	}

	/** Gets the output stream for the message type. */
	public int getOutputLevel(String messageType) {
		StreamLevelPair pair = ((StreamLevelPair)logStreams.get(messageType));
		if (pair != null) {
			return pair.outputLevel;
		} else {
			return defaultLog.outputLevel;
		}
	}

	/** Prints a message on the default log stream */
	final public void logOutput(String message) {
		if (1 <= getOutputLevel("log")) {
			getLogStream("log").println("++ log: " + message);
		}
	}

	/** Prints a message of type messageType on the log stream */
	final public void logOutput(String messageType, String message) {
		if (1 <= getOutputLevel(messageType)) {
			getLogStream(messageType).println("++ " + messageType + " log: " + message);
		}
	}

	/** Prints a message on the default error stream */
	final public void logErrorOutput(String message) {
		if (1 <= getOutputLevel("error")) {
			getLogStream("error").println("!! " + "error" + ": " + message);
		}
	}

	/** Prints a message of type messageType on the error stream */
	final public void logErrorOutput(String messageType, String message) {
		if (1 <= getOutputLevel(messageType)) {
			getLogStream(messageType).println("!! " + messageType + " error: " + message);
		}
	}

	/** Prints a message on the default log stream */
	final public void logDebugOutput(String message) {
		if (1 <= getOutputLevel("debug")) {
			getLogStream("debug").println(" -- " + "debug" + ": " + message);
		}
	}

	/** Prints a message of type messageType on the debug stream */
	final public void logDebugOutput(String messageType, String message) {
		if (1 <= getOutputLevel(messageType)) {
			getLogStream(messageType).println(" -- " + messageType + " debug: " + message);
		}
	}

	/** Prints a message on the default debug stream */
	final public void logDebugOutput(String message, int debugLevel) {
		if (debugLevel <= getOutputLevel("debug")) {
			getLogStream("debug").println(" -- " + "debug" + ": " + message);
		}
	}

	/** Prints a message of type messageType on the debug stream */
	final public void logDebugOutput(String messageType, String message, int debugLevel) {
		if (debugLevel <= getOutputLevel(messageType)) {
			getLogStream(messageType).println(" -- " + messageType + " debug: " + message);
		}
	}

} // end LogOutput ---------------------------------------------------------------------------------
