package mosaic.util.custom;

import java.beans.*;
import java.awt.Component;
import java.awt.event.*;
import java.awt.AWTEvent;
import java.io.Serializable;
import java.util.EventObject;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import mosaic.util.SwingInformer;

/**
 * CellEditor, which uses a PropertyEditor and tries to display the value
 * by calling its getAsText().
 *
 * Uses a parts of Suns code of DefaultCellEditor
 *
 * changes history:
 * 15.11.1999   Hilpold fix bug: last change in array not propagated
 *                               fixed by changing CustomArrayPropEditor
 *
 *
 *
 * @version 15 Nov 1999
 * @author Thomas Hilpold
 * @see CustomArrayCellEditor
 */
public class CustomCellTextEditor implements TableCellEditor, TreeCellEditor,
					  Serializable {

    //
    //  Instance Variables
    //

    /** Event listeners */
    protected EventListenerList listenerList = new EventListenerList();
    transient protected ChangeEvent changeEvent = null;

    protected JComponent editorComponent;
    protected EditorDelegate delegate;
    protected int clickCountToStart = 1;
    protected boolean supportsGetAsText;
    protected PropertyEditor pe;
    //
    // Debugging
    //
    boolean[] DEBUG = new boolean[]{false};

//
//  Constructors
//

    public CustomCellTextEditor(PropertyEditor aPe) {

        String init;

        this.pe = aPe;
        this.editorComponent = new JTextField();
    	this.clickCountToStart = 2;
        init = pe.getAsText();
        supportsGetAsText = (init != null);
        this.delegate = new EditorDelegate() {
            public void setValue(Object x) {
                if (DEBUG[0]) System.err.println("setValue: " + x);
       		    //editorComponent.requestFocus();
                super.setValue(x);
                if (pe.getValue() != x) pe.setValue(x);
	        	if (x != null)
        		    ((JTextField)editorComponent).setText(pe.getAsText());
        		else
        		    ((JTextField)editorComponent).setText("");

           }

	        public Object getCellEditorValue() {
                String text = ((JTextField)editorComponent).getText();
                if (DEBUG[0]) System.err.println("getCEValue: " + text);
        		return text;
    	    }

	        public boolean startCellEditing(EventObject anEvent) {
                if (DEBUG[0]) System.err.println("startCellEditing by "
                                                 + "requesting focus");
       		    //editorComponent.requestFocus();
        		return true;
    	    }

	        public boolean stopCellEditing() {
                if (DEBUG[0]) System.err.println("StopCellEditing by "
                                                 + "doing nothing");
        		return true;
    	    }

	        public void cancelCellEditing() {
                if (DEBUG[0]) System.err.println("CancelCellEditing by "
                                                 + "doing nothing");
    	    }
        };
    }

    public Component getComponent() {
        if (DEBUG[0]) System.err.println("getComponent: " + editorComponent);
    	return editorComponent;
    }

//
//  Modifying
//

    public void setClickCountToStart(int count) {
    	clickCountToStart = count;
    }

    /**
     *  clickCountToStart controls the number of clicks required to start
     *  editing if the event passed to isCellEditable() or startCellEditing() is
     *  a MouseEvent.  For example, by default the clickCountToStart for
     *  a JTextField is set to 2, so in a JTable the user will need to
     *  double click to begin editing a cell.
     */
    public int getClickCountToStart() {
    	return clickCountToStart;
    }

//
//  Implementing the CellEditor Interface
//

    public Object getCellEditorValue() {
        return delegate.getCellEditorValue();
    }

    public boolean isCellEditable(EventObject anEvent) {
	if ((anEvent != null) && (anEvent instanceof MouseEvent)) {
	    if (((MouseEvent)anEvent).getClickCount() < clickCountToStart)
		return false;
	}
	return delegate.isCellEditable(anEvent);
    }

    public boolean shouldSelectCell(EventObject anEvent) {
	boolean         retValue = true;

	if (this.isCellEditable(anEvent)) {
	    if (anEvent == null || ((MouseEvent)anEvent).getClickCount() >=
		clickCountToStart)
		retValue = delegate.startCellEditing(anEvent);
	}

	// By default we want the cell the be selected so
	// we return true
	return retValue;
    }

    public boolean stopCellEditing() {
    	boolean stopped = delegate.stopCellEditing();
    	if (stopped) {
    	    fireEditingStopped();
    	}
    	return stopped;
    }

    public void cancelCellEditing() {
    	delegate.cancelCellEditing();
    	fireEditingCanceled();
    }

//
//  Handle the event listener bookkeeping
//
    public void addCellEditorListener(CellEditorListener l) {
    	listenerList.add(CellEditorListener.class, l);
    }

    public void removeCellEditorListener(CellEditorListener l) {
    	listenerList.remove(CellEditorListener.class, l);
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireEditingStopped() {
        // Guaranteed to return a non-null array
        Object[] listeners = listenerList.getListenerList();
        // Process the listeners last to first, notifying
        // those that are interested in this event
        for (int i = listeners.length-2; i>=0; i-=2) {
            if (listeners[i]==CellEditorListener.class) {
            // Lazily create the event:
            if (changeEvent == null)
                changeEvent = new ChangeEvent(this);
            ((CellEditorListener)listeners[i+1]).editingStopped(changeEvent);
            }
        }
    }

    /*
     * Notify all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.
     * @see EventListenerList
     */
    protected void fireEditingCanceled() {
        // Guaranteed to return a non-null array
        Object[] listeners = listenerList.getListenerList();
        // Process the listeners last to first, notifying
        // those that are interested in this event
        for (int i = listeners.length-2; i>=0; i-=2) {
            if (listeners[i]==CellEditorListener.class) {
            // Lazily create the event:
            if (changeEvent == null)
                changeEvent = new ChangeEvent(this);
            ((CellEditorListener)listeners[i+1]).editingCanceled(changeEvent);
            }
        }
    }

//
//  Implementing the TreeCellEditor Interface
//

    public Component getTreeCellEditorComponent(JTree tree, Object value,
						boolean isSelected,
						boolean expanded,
						boolean leaf, int row) {
	String         stringValue = tree.convertValueToText(value, isSelected,
					    expanded, leaf, row, false);

	delegate.setValue(stringValue);
	return editorComponent;
    }

//
//  Implementing the CellEditor Interface
//

    public Component getTableCellEditorComponent(JTable table, Object value,
						 boolean isSelected,
						 int row, int column) {
    delegate.setValue(value);
	return editorComponent;
    }


//
//  Protected EditorDelegate class
//

    protected class EditorDelegate
            implements Serializable {

        protected Object value;

        public Object getCellEditorValue() {
            return value;
        }

        public void setValue(Object x) {
            this.value = x;
        }

        public boolean isCellEditable(EventObject anEvent) {
	        return true;
    	}

        public boolean startCellEditing(EventObject anEvent) {
    	    return true;
    	}

        public boolean stopCellEditing() {
	        return true;
    	}

        public void cancelCellEditing() {
    	}
    } // class  EditorDelegate
}
