package mosaic.util.custom;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.lang.reflect.Method;
import javax.swing.*;
import javax.swing.border.*;

/**
 *  A PropertySheet for a Beans properties that analyzes the Bean.
 *  The original object newBean will remain unchanged.
 *  Changes will be applied to its deepcopy and written back to the
 *  propertyeditor after each propertychange
 *  <p>
 *  Currently no support for IndexedPropertyDescriptors
 *  and neither for PropertyEditors, which support paintValue().
 *  You can use normal PropertyDescs for indexed Props especially with JBuilder.
 *  Beans must be serializable.
 *  Uses AWT choice - all other SWING Components
 *  <p>
 *  This class is intended to be used by the PropertyCustomEditor class, which
 *  in turn is a customEditor for GenericBeanEditor.
 *
 *  First release date:
 *  09 Jan 1999
 *
 *  Changes:
 *  01 Mar 1999 Hilpold | Gui uses class sim.custom.ScrollablePanel to
 *                      | make the propPanels Xsize dependent on the available
 *                      | viewport (no horizontal scrollbar)
 *
 *  @author Thomas Hilpold
 *  @version 01 Mar 1999
 *  @see PropertyCustomEditor
 *  @see GenericBeanEditor
 */
public class PropertySheet extends Object
    implements ActionListener, ItemListener, FocusListener {
    //Methods
    public PropertySheet(Object bean, PropertySheetListener myListener) {
        this.myListener = myListener;
        beanChanged(bean);
    }

    /**
     *  Deepcopies newBean using Serialization, analyzes it and updates
     *  the view.
     */
    public void beanChanged(Object newBean) {
        try {
            //deep copy Bean using Serialization
            ByteArrayOutputStream bOut = new ByteArrayOutputStream();
            ObjectOutputStream out     = new ObjectOutputStream(bOut);
            out.writeObject(newBean);
            ByteArrayInputStream bIn =  new ByteArrayInputStream(bOut.toByteArray());
            ObjectInputStream in     = new ObjectInputStream(bIn);
            bean = in.readObject();
        } catch (Exception e) {
           System.err.println(this +  e.toString());
           System.err.println("New Bean  = " + newBean);
           System.err.println("Beanclass = " + newBean.getClass());
           System.err.println("uncopied bean will be used - changes can be lost");
           bean = newBean;
        };
        initialize();
        analyzeBean();
        // Scrollablepanels xsize will depend on viewport
        JPanel p = new ScrollablePanel();
        GridBagConstraints propPanelConstr = new GridBagConstraints();
        propPanelConstr.weightx = 1;
        propPanelConstr.weighty = 1;
        propPanelConstr.fill = GridBagConstraints.HORIZONTAL;
        propPanelConstr.anchor = GridBagConstraints.NORTHWEST;
        p.setLayout(new GridBagLayout());
        // add propPanel with topleft alignment, even if viewport > propPanel
        p.add(propPanel, propPanelConstr);
        scrollPane = new JScrollPane();
        scrollPane.setHorizontalScrollBarPolicy(
            JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scrollPane.setVerticalScrollBarPolicy(
            JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        scrollPane.getViewport().setBackingStoreEnabled(true);
        //scrollPane.getViewport().setPreferredSize(propPanel.getPreferredSize());
        scrollPane.getViewport().setView(p);
    }

    /**
     * Handles Actionsevents sent by Buttons and JTextfields
     * Buttons: Start CustomEditorDialog
     * JTextFields: exist for PropertyEditors wich support getAsText()
     * but have no Tags and no CustomEditor.
     */
    public void actionPerformed(ActionEvent e){
        //Registered are - Buttons to start CustomEditors
        //               - TextFields to edit values as Strings
        if (e.getSource() instanceof JButton) {
            //startCustomEditor
            JButton source = (JButton)e.getSource();
            int i = 0;
            boolean found = false;
            while ((i < numOfProps) && (!found)) {
                found = (source == propCustomStartButtons[i]);
                if (!found) i++;
            }
            if (found) {
                if ((propEditors[i] != null)
                    && (propEditors[i].supportsCustomEditor())) {
                    JDialog cusDlg = new CustomEditorDialog(
                        propEditors[i].getCustomEditor(),
                        "PropertySheet : " + propDescs[i].getName()
                        );
                    cusDlg.setLocationRelativeTo(source);
                    cusDlg.setVisible(true);
                    scrollPane.repaint();
                    //modal dlg open
                    updateBean(i);
                }
            }
        }
        else if (e.getSource() instanceof JTextField) {
            JComponent source = (JComponent)e.getSource();
            int i = 0;
            boolean found = false;
            while ((i < numOfProps) && (!found)) {
                found = (source == propValueComps[i]);
                if (!found) i++;
            }
            if (found) {
                try {
                    propEditors[i].setAsText(((JTextField)source).getText());
                } catch (IllegalArgumentException ex) {
                    JOptionPane.showMessageDialog(null,
                        "Could not set Value !\n" +
                        ((ex.getMessage() != null)? ex.getMessage():""));
                }
                ((JTextField)source).setText(propEditors[i].getAsText());
                updateBean(i);
                source.transferFocus();
            }
        }
    }
    public void focusGained(FocusEvent e){
        // do nothing
    }

    /**
     * Invoked when a component loses the keyboard focus.
     */
    public void focusLost(FocusEvent e) {
        if (e.getSource() instanceof JTextField) {
            JComponent source = (JComponent)e.getSource();
            int i = 0;
            boolean found = false;
            while ((i < numOfProps) && (!found)) {
                found = (source == propValueComps[i]);
                if (!found) i++;
            }
            if (found) {
                try {
                    propEditors[i].setAsText(((JTextField)source).getText());
                } catch (IllegalArgumentException ex) {
                    JOptionPane.showMessageDialog(null,
                        "Could not set Value !\n" +
                        ((ex.getMessage() != null)? ex.getMessage():""));
                }
                ((JTextField)source).setText(propEditors[i].getAsText());
                updateBean(i);
            }
        }
    }

    /**
     * called by Choices of Properties whose PropertyEditor supports Tags.
     * Tries a setAsText(chosenString) on the corresponding PropertyEditor
     * to update it.
     */
    public void itemStateChanged(ItemEvent e) {
        //registered are only Choices for tagged PEs;
        boolean found = false;
        int i = 0;
        Component source = (Component) e.getSource();
        while ((!found) && (i < numOfProps)) {
            found = (source == propValueComps[i]);
            if (!found) i++;
        }
        if (found) {
            try {
                propEditors[i].setAsText((String)((Choice)source).getSelectedItem());
            } catch (IllegalArgumentException ex) {
                //should never happen !
                JOptionPane.showMessageDialog(null,
                        "Could not set Value from Choice !\n" +
                        ((ex.getMessage() != null)? ex.getMessage():""));
                ((Choice)source).select(propEditors[i].getAsText());
            }
            updateBean(i);
        }
    }

    /** simply creates a new JPanel to put the whole stuff in */
    protected void initialize(){
        propPanel = new ScrollablePanel();
        propPanel.setLayout(new GridBagLayout());
    }

    /** returns the scrollpane to be added to another container */
    public JScrollPane getScrollPane() {
        return scrollPane;
    }

    /**
     * Analyzes the current Bean with the Introspector.
     * Searches for PropertyEditors for each Property by:
     * 1. asking the PropertyDescriptor if it has one
     * 2. asking PropertyEditorManager, if it can find one
     *      this in turn will a) look in its registry
     *                        b) try to look for a class ptype+"Editor" in the
     *                           package of ptype.
     *                        c) look for a class ptype+"Editor" on a special
     *                           searchpath.
     * Currently no support for IndexedPropertyDescriptors.
     */
    protected void analyzeBean(){
       Object curValue;
       Object[] args =new Object[0];
       if (bean != null) {
         try {
             propDescs = Introspector.getBeanInfo(bean.getClass()).getPropertyDescriptors();
          } catch (Exception e) {
              propDescs = null;
              System.out.println("No PropertyDescriptors");
          };
          if (propDescs != null) {
              PropertyDescSorter.sortAlpha(propDescs);
              numOfProps = propDescs.length;
              //initialize Arrays except propDescs
              propNameLabels = new JLabel[numOfProps];
              propValueComps = new Component[numOfProps];
              propCustomStartButtons = new JButton[numOfProps];
              propEditors = new PropertyEditor[numOfProps];
              //
              for (int i = 0; i < numOfProps; i++) {
                  //ask propDesc if explicit Editor used
                  if (propDescs[i].getPropertyEditorClass() != null) {
                      try{
                        propEditors[i] =
                          ((PropertyEditor)propDescs[i].getPropertyEditorClass().newInstance());
                      }catch(Exception e) {
                          System.err.println(e.toString());
                          propEditors[i] = null;
                      }
                  }
                  else {
                      //
                      // PENDING NO SUPPORT FOR INDEXEDPROPERTYDESCS
                      // THEY RETURN NULL FOR propertyType
                      //
                      Class propType;
                      propType = propDescs[i].getPropertyType();
                      if (propType != null) {
                          propEditors[i] =
                            PropertyEditorManager.findEditor(propType);
                      }
                  }
                  try {
                      curValue = propDescs[i].getReadMethod().invoke(bean,args);
                  }catch (Exception e) {
                    System.err.println("" + this + " noreadmethod or no access");
                    curValue = null; /* !readAble*/
                  };

                  /**
                   * some propEditors are not robust enough to allow
                   * null to be set (eg sun.beans.ColorEditor)
                   */
                  if ((propEditors[i] != null) && (curValue != null)) {
                      propEditors[i].setValue(curValue);
                      //System.out.println("Propvalset:"+propEditors[i].getAsText());
                  }
                  buildGUIFor(i, curValue);
              } //end for
          }
       }
    }

    /**
     * creates an appropriate GUI component for a given Property and sets the
     * value.
     * <p>
     * Which component will be created is decided depending on the Property-
     * Editor as follows:
     * <p>
     *      noPropEd or noSetterMethod  -> JLabel
     *      supportsTags                -> Choice (AWT)
     *      supportsCustomEditor        -> JLabel with Button to call customEd
     *      noTags && noCustomEd        -> JTextfield
     *      other                       -> Error
     */
    private void buildGUIFor(int propNum, Object curValue){
        //curValue only needed, when there is no PropertyEditor
        //build a label for the propertys name
        GridBagConstraints left, right;
        boolean isWriteable, needUpdate;
        left = new GridBagConstraints();
        right = new GridBagConstraints();
        left.gridx = 0;
        right.gridx = 1;
        left.gridy = propNum;
        right.gridy =left.gridy;
        right.weightx = 1;
        left.weightx = 0;
        left.insets = new Insets(0,5,0,5);
        left.fill = GridBagConstraints.HORIZONTAL;
        right.fill = GridBagConstraints.BOTH;
        needUpdate = true;
        isWriteable = (propDescs[propNum].getWriteMethod() !=  null);
        propNameLabels[propNum] =
            new JLabel(propDescs[propNum].getName()+" : ", SwingConstants.RIGHT);
        propNameLabels[propNum].setMinimumSize(
            new Dimension(propNameLabels[propNum].getPreferredSize().width, 20));
        propNameLabels[propNum].setPreferredSize(
            new Dimension(propNameLabels[propNum].getPreferredSize().width, 20));
        propPanel.add(propNameLabels[propNum], left);
        if ((propEditors[propNum] == null) || (!isWriteable)) {
            //build only label to display the props Value
            JLabel vLabel = new JLabel("-----", SwingConstants.RIGHT);
            propPanel.add(vLabel, right);
            propValueComps[propNum]= vLabel;
            }
        else if ((propEditors[propNum].getTags() != null)
                  && (!propEditors[propNum].supportsCustomEditor())) {
            //build a Choice
            propEditors[propNum].setValue(curValue);
            String[] pTags = propEditors[propNum].getTags();
            Choice   c = new Choice();
            c.setFont(new Font("System",1,9));
            c.setSize(c.getSize().width, 20);
            for (int j = 0; j < pTags.length; j++) c.addItem(pTags[j]);
            c.addItemListener(this);
            //PENDING setSelectedItem
            propPanel.add(c, right);
            propValueComps[propNum] = c;
            }
        else if (propEditors[propNum].supportsCustomEditor()) {
            //build a Label
            JPanel customPanel = new JPanel(new BorderLayout());
            JLabel vLabel = new JLabel("-----", SwingConstants.LEFT);
            customPanel.add(vLabel, "Center");
            propValueComps[propNum]= vLabel;
            //build a button register this as actionListener
            JButton ceButton = new JButton("...");
            ceButton.addActionListener(this);
            customPanel.add(ceButton, "East");
            customPanel.setMinimumSize(new Dimension(20, 20));
            propPanel.add(customPanel, right);
            propCustomStartButtons[propNum] = ceButton;
            }
        else if ((propEditors[propNum].getTags() == null)
                  && (!propEditors[propNum].supportsCustomEditor())) {
            //build a TextField for String Editing register ActionListener
            JTextField t = new JTextField("-----");
            t.setHorizontalAlignment(SwingConstants.RIGHT);
            t.addActionListener(this);
            t.addFocusListener(this);
            t.setMinimumSize(new Dimension(20, 20));
            propPanel.add(t, right);
            propValueComps[propNum] = t;
            }
        else {
            System.err.println("Could not determine how to edit "+propDescs[propNum].getName());
            needUpdate = false;
            }
        if (needUpdate) updateComponent(propNum, curValue);
    }

    /**
     * Gets the value from a PropertyEditor and appplies it to the bean.
     * Uses Bean API to determine and Reflection API to
     * invoke the setter Method.
     * <p>
     * After a set on a bean all properties have to be read back in to
     * prevent the danger of indirect property changes.
     */
    private void updateBean(int propNum) {
        //Updates the Bean with information from the PropertyEditor
        //if such exists
        Method setter;
        PropertyEditor editor;
        setter = propDescs[propNum].getWriteMethod();
        editor = propEditors[propNum];
        if ((setter != null) && (editor != null)) {
            Object[] parm = new Object[1];
            parm[0] = editor.getValue();
            try {
                setter.invoke(bean, parm);
            }
            catch (Exception e) {System.out.println("upDateBean"+ e + " . " + parm[0]);};
            System.out.println("updateBean" + parm[0].toString());
            //other properties could have changed,
            // so we need to read all vals from the Bean
            // and update the PropEditors
            updateGUI();
        }
        else
            System.out.println("updateBean - could not apply"+editor);
    }

    /** applies the values of the beans properties to the GUI components. */
    private void updateGUI() {
        //read bean properties -> write them to the editors
        //apply changes to GUI Components
        Object[] parm = new Object[0];
        Object curValue;
        Method curReader;
        PropertyEditor curEditor;
        for (int i = 0; i < numOfProps; i++) {
            curReader = propDescs[i].getReadMethod();
            curEditor = propEditors[i];
            if (curReader != null) {
                try {
                    curValue = curReader.invoke(bean, parm);
                }
                catch (Exception e) {
                    curValue = null;
                    System.out.println("updateEditors"+ e);
                };
            }
            else curValue = null;
            if ((curEditor != null) && (curEditor.getValue() != curValue)) {
                curEditor.setValue(curValue);
            }
            updateComponent(i, curValue);
        }
        //calcAndSetPropPrefSize();
    }

    /** applies a given value to the Gui component of the given property */
    private void updateComponent(int propNum, Object value) {
        //value is used if no Editor exists
        Component c = propValueComps[propNum];
        //
        boolean isReadable  = (propDescs[propNum].getReadMethod() !=  null);
        if (c instanceof JLabel) {
            if (value != null) {
                //try to use getAsText if a PropertyEditor exists
                if (propEditors[ propNum] != null) {
                    String newVal = propEditors[propNum].getAsText();
                    if (newVal != null)
                        ((JLabel)c).setText(newVal);
                    else
                        ((JLabel)c).setText(value.toString());
                }
                else
                   ((JLabel)c).setText(value.toString());
            }
            else
                ((JLabel)c).setText("null");
        }
        else if (c instanceof Choice){ //AWT Choice used
            ((Choice)c).removeItemListener(this);
            ((Choice)c).select(propEditors[propNum].getAsText());
            ((Choice)c).addItemListener(this);
        }
        else if (c instanceof JTextField){
            ((JTextField)c).removeActionListener(this);
            if (isReadable) {
                ((JTextField)c).setText(propEditors[propNum].getAsText());
                }
            else
                ((JTextField)c).setText("notreadable");
            ((JTextField)c).addActionListener(this);
        }
        else {
            System.out.println("updateComponent: no valid Component");
        }
        //notify the only Listener that something changed
        //PENDING
        if (myListener != null) {
            myListener.fireChanges(propDescs[propNum].getName());
        }
        //propPanel.doLayout();
    }

    /**
     * returns the current Bean.
     * Usually called to write it back to the PropertyEditor of the whole Bean
     */
    public Object getBean() {
      return bean;
    }

    //
    //Data
    //

    protected Object bean;
    protected JPanel propPanel;
    protected JScrollPane scrollPane;
    protected JLabel[] propNameLabels;
    protected Component[] propValueComps;
    protected JButton[] propCustomStartButtons;
    protected PropertyDescriptor[] propDescs;
    protected PropertyEditor[] propEditors;
    protected int numOfProps;
    protected PropertySheetListener myListener;
}



