package sim.util;

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * The class LogPrinter provides facilities to easily write to one or more Logfiles via
 * static methods. It will always be appended to exising Files, but Methods
 * to delete one or all of them before writing to them are supported.
 *
 * The main method shows an example.
 *
 * The Logfile will be named:
 * baseName + "1"
 * baseName + "2"
 *
 * This is Thread safe, but blocking may occur during concurrent access.
 *
 * Please report bugs or suggestions to hilpold@swe.uni-linz.ac.at
 *
 * @author Thomas Hilpold
 * @version 15.07.1999
 **/
public class HighPerformanceCounter extends Object {

    long startTickH;
    long startTickL;
    long ticksPerSecond;
    boolean available;

    static{
        try {
            System.loadLibrary("HighPerformanceCounter");
        } catch (Exception e) {
            System.err.println("NO HighPerformanceCounter avaiable");
        }
    }

    public HighPerformanceCounter() {
        ticksPerSecond = getTicksPerSecond();
        available = (ticksPerSecond != 0);
        if (!available) {
            System.err.println("NO HighPerformanceCounter avaiable");
        }
        startTickH = 0;
        startTickL = 0;
    }

    public void resetTime() {
        startTickH = getTickCountH();
        startTickL = getTickCountL();
    }

    public double getTime() {
        long ticksPassedH;
        long ticksPassedL;

        ticksPassedH = getTickCountH() - startTickH;
        ticksPassedL = getTickCountL() - startTickL;
        if (ticksPassedL < 0) {
            ticksPassedL += Long.MAX_VALUE;
        }

        if (ticksPassedH < 0) {
            ticksPassedH += Long.MAX_VALUE;
        }

        return ticksPassedL / (double)ticksPerSecond * 1000;
    }

    public double getTicksPerMs() {
        return ticksPerSecond / 1000.0;
    }

    public double calcMaxDurationInSecs() {
        return Long.MAX_VALUE / getTicksPerSecond();
    }

    private native long getTicksPerSecond();
    private native long getTickCountH();
    private native long getTickCountL();

    public static void main(String[] argv) {
        HighPerformanceCounter x = new HighPerformanceCounter();
        System.out.println("FMax " + Float.MAX_VALUE);
        System.out.println("FMin " + Float.MIN_VALUE);
        System.out.println("Ticks/ms : " + x.getTicksPerMs());
        System.out.println("Ticks/ms : " + x.getTicksPerMs());
        System.out.println("reset : ");
        x.resetTime();
        System.out.println("time : " + x.getTime());
        System.out.println("MaxDur : " + x.calcMaxDurationInSecs() / 60 / 60 + "hours");
        System.out.println("LongMax : " + Long.MAX_VALUE);
        for (int f = 1; f <= 100; f++) {
            try {
                Thread.sleep(1000);
            } catch (Exception e) {
            }
            System.out.println("time after 1000 sleep : " + x.getTime());
            System.out.println("time immediatly after: " + x.getTime());
            System.out.println("real TickcountL: " + x.getTickCountL());
        }
    }
}

